import logging
import os
from typing import Any

import openai
import pytest
from psycopg import Connection
from psycopg.rows import dict_row

from tests.vectorizer import expected
from tests.vectorizer.cli.conftest import (
    TestDatabase,
    configure_vectorizer,
    run_vectorizer_worker,
    setup_source_table,
)


def configure_openai_vectorizer(
    connection: Connection,
    openai_proxy_url: str | None = None,
    number_of_rows: int = 1,
    concurrency: int = 1,
    batch_size: int = 1,
    chunking: str = "chunking_character_text_splitter()",
    formatting: str = "formatting_python_template('$chunk')",
) -> int:
    """Creates and configures a vectorizer for testing"""
    table_name = setup_source_table(connection, number_of_rows)
    base_url = (
        f", base_url => '{openai_proxy_url}'" if openai_proxy_url is not None else ""
    )

    embedding = f"embedding_openai('text-embedding-ada-002', 1536{base_url})"
    return configure_vectorizer(
        table_name,
        connection,
        concurrency=concurrency,
        batch_size=batch_size,
        chunking=chunking,
        formatting=formatting,
        embedding=embedding,
    )


@pytest.mark.parametrize(
    "num_items,concurrency,batch_size,openai_proxy_url,secrets_from_db",
    [
        (1, 1, 1, None, False),
        (1, 1, 1, 8000, False),
        (4, 2, 2, None, False),
        (1, 1, 1, None, True),
        (4, 2, 2, None, True),
    ],
    indirect=["openai_proxy_url"],
)
def test_process_vectorizer(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
    num_items: int,
    concurrency: int,
    batch_size: int,
    openai_proxy_url: str | None,
    secrets_from_db: bool,
):
    """Test successful processing of vectorizer tasks"""
    _, conn = cli_db
    vectorizer_id = configure_openai_vectorizer(
        cli_db[1],
        openai_proxy_url,
        number_of_rows=num_items,
        concurrency=concurrency,
        batch_size=batch_size,
    )
    # Insert pre-existing embedding for first item
    with conn.cursor() as cur:
        cur.execute("""
           INSERT INTO
           blog_embedding_store(embedding_uuid, id, chunk_seq, chunk, embedding)
           VALUES (gen_random_uuid(), 1, 1, 'post_1',
            array_fill(0, ARRAY[1536])::vector)
        """)

    if secrets_from_db:
        # create extension to test loading secret from db
        conn.execute("""CREATE EXTENSION IF NOT EXISTS ai CASCADE""")
        # Ensuring no OPENAI_API_KEY env set for the worker
        # to test loading secret from db
        del os.environ["OPENAI_API_KEY"]

    # When running the worker with cassette matching original test params
    cassette = (
        f"openai-character_text_splitter-chunk_value-"
        f"items={num_items}-batch_size={batch_size}-"
        f"custom_base_url={openai_proxy_url is not None}.yaml"
    )
    # logging.getLogger("vcr").setLevel(logging.DEBUG)

    with vcr_.use_cassette(cassette):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id, concurrency)

    print(f"result: {result.stdout}")
    assert not result.exception
    assert result.exit_code == 0

    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) as count FROM blog_embedding_store;")
        assert cur.fetchone()["count"] == num_items  # type: ignore

    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM ai.vectorizer_worker_process;")
        row = cur.fetchone()
        assert row is not None
        assert row["started"] is not None
        assert row["last_heartbeat"] is not None
        assert row["heartbeat_count"] >= 1
        assert row["error_count"] == 0
        assert row["last_error_at"] is None
        assert row["last_error_message"] is None
        assert row["success_count"] == num_items
        worker_id = row["id"]
        assert cur.fetchone() is None

        cur.execute("SELECT * FROM ai.vectorizer_worker_progress;")
        row = cur.fetchone()
        assert row is not None
        assert row["last_success_at"] is not None
        assert row["last_success_process_id"] == worker_id
        assert row["last_error_at"] is None
        assert row["last_error_message"] is None
        assert row["last_error_process_id"] is None
        assert row["success_count"] == num_items
        assert cur.fetchone() is None


@pytest.mark.postgres_params(load_openai_key=False)
def test_vectorizer_without_secrets_fails(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    _, conn = cli_db
    vectorizer_id = configure_openai_vectorizer(cli_db[1])
    # Insert pre-existing embedding for first item
    with conn.cursor() as cur:
        cur.execute("""
           INSERT INTO
           blog_embedding_store(embedding_uuid, id, chunk_seq, chunk, embedding)
           VALUES (gen_random_uuid(), 1, 1, 'post_1',
            array_fill(0, ARRAY[1536])::vector)
                """)
    # Ensuring no OPENAI_API_KEY env set for the worker
    del os.environ["OPENAI_API_KEY"]

    cassette = "openai-character_text_splitter-chunk_value-items=1-batch_size=1.yaml"
    logging.getLogger("vcr").setLevel(logging.DEBUG)
    with vcr_.use_cassette(cassette):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    assert result.exit_code == 1
    assert "ApiKeyNotFoundError" in result.output
    print(f"result: {result.stdout}")

    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM ai.vectorizer_worker_process;")
        row = cur.fetchone()
        assert row is not None
        assert row["started"] is not None
        assert row["last_heartbeat"] is not None
        assert row["heartbeat_count"] >= 1
        assert row["error_count"] == 1
        assert row["success_count"] == 0
        assert row["last_error_at"] is not None
        assert row["last_error_message"] is not None
        worker_id = row["id"]
        assert cur.fetchone() is None

        cur.execute("SELECT * FROM ai.vectorizer_worker_progress;")
        row = cur.fetchone()
        assert row is not None
        assert row["last_success_at"] is None
        assert row["last_success_process_id"] is None
        assert row["last_error_at"] is not None
        assert row["last_error_message"] is not None
        assert row["last_error_process_id"] == worker_id
        assert row["success_count"] == 0
        assert cur.fetchone() is None


def test_document_exceeds_model_context_length(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test handling of documents that exceed the model's token limit"""
    _, conn = cli_db
    # Given a vectorizer configuration
    vectorizer_id = configure_openai_vectorizer(
        cli_db[1],
        number_of_rows=2,
        batch_size=2,
        chunking="chunking_recursive_character_text_splitter(128, 10,"
        " separators => array[E'\n\n'])",
    )
    with conn.cursor(row_factory=dict_row) as cur:
        long_content = "AGI" * 5000
        cur.execute(
            f"UPDATE blog SET CONTENT = '{long_content}' where id = '2'",
        )

    # When running the worker
    with vcr_.use_cassette("test_document_in_batch_too_long.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    assert result.exit_code == 0

    # Then only the normal document should be embedded
    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM blog_embedding_store ORDER BY id")
        records = cur.fetchall()
        assert len(records) == 2
        record = records[0]

        # Verify the embedded document
        assert record["id"] == 1
        assert record["chunk"] == "post_1"
        assert (
            record["embedding"]
            == expected.embeddings["openai-character_text_splitter-chunk_value-1-1"][0]
        )

        record = records[1]

        assert record["id"] == 2
        assert record["chunk"] == long_content[:15001]
        assert (
            record["embedding"]
            == expected.embeddings["openai-character_text_splitter-chunk_value-1-1"][1]
        )

        # Check error was logged
        cur.execute("SELECT * FROM ai.vectorizer_errors")
        errors = cur.fetchall()
        assert len(errors) == 0


def test_invalid_api_key_error(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test that worker handles invalid API key appropriately"""
    _, conn = cli_db

    vectorizer_id = configure_openai_vectorizer(
        cli_db[1],
        number_of_rows=2,
        batch_size=2,
        chunking="chunking_recursive_character_text_splitter()",
    )

    os.environ["OPENAI_API_KEY"] = "invalid"

    # When running the worker and getting an invalid api key response
    with vcr_.use_cassette("test_invalid_api_key_error.yaml"):
        try:
            run_vectorizer_worker(cli_db_url, vectorizer_id)
        except openai.AuthenticationError as e:
            assert e.code == 401

    # Ensure there's an entry in the errors table
    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM ai.vectorizer_errors")
        records = cur.fetchall()
        assert len(records) == 1
        error = records[0]
        assert error["id"] == vectorizer_id
        assert error["message"] == "embedding provider failed"
        assert error["details"] == {
            "provider": "openai",
            "error_reason": "Error code: 401 - {'error': {'message': "
            "'Incorrect API key provided: invalid."
            " You can find your API key at"
            " https://platform.openai.com/account/api-keys.',"
            " 'type': 'invalid_request_error', 'param': None,"
            " 'code': 'invalid_api_key'}}",
        }


def test_invalid_function_arguments(
    cli_db: tuple[TestDatabase, Connection], cli_db_url: str
):
    """Test that worker handles invalid embedding model arguments appropriately"""
    _, conn = cli_db

    vectorizer_id = configure_openai_vectorizer(cli_db[1])
    # And a vectorizer with invalid embedding dimensions
    with conn.cursor() as cur:
        cur.execute(
            """
            UPDATE ai.vectorizer
            SET config = jsonb_set(
                config,
                '{embedding,dimensions}',
                '128'::jsonb
            )
            WHERE id = %s
        """,
            (vectorizer_id,),
        )

    # When running the worker
    try:
        run_vectorizer_worker(cli_db_url, vectorizer_id)
    except ValueError as e:
        assert str(e) == "dimensions must be 1536 for text-embedding-ada-002"

    # Then an error was logged
    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT * FROM ai.vectorizer_errors")
        records = cur.fetchall()
        assert len(records) == 1
        error = records[0]
        assert error["id"] == vectorizer_id
        assert error["message"] == "embedding provider failed"
        assert error["details"] == {
            "provider": "openai",
            "error_reason": "dimensions must be 1536 for text-embedding-ada-002",
        }



"""Models for representing data being vectorized in the semantic catalog.

This module contains the data models used for representing content that will be
converted to vector embeddings for the semantic catalog.
"""

from collections.abc import Sequence

from pydantic import BaseModel


class EmbedRow(BaseModel):
    """Model representing a row of data to be embedded.

    This class holds the content to be embedded and its resulting vector embedding.
    It serves as the data structure for passing content to embedding services and
    storing the resulting vector embeddings.

    Attributes:
        id: Database ID of the item being embedded.
        content: Text content to be converted to a vector embedding.
        vector: The resulting vector embedding (None until embedding is performed).
    """

    id: int
    content: str
    vector: Sequence[float] | None = None



from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import override

from pgai.vectorizer.chunking import (
    Chunker,
    LangChainCharacterTextSplitter,
    LangChainRecursiveCharacterTextSplitter,
)
from pgai.vectorizer.embedders import LiteLLM, Ollama, OpenAI, VoyageAI
from pgai.vectorizer.formatting import ChunkValue, PythonTemplate
from pgai.vectorizer.loading import ColumnLoading
from pgai.vectorizer.parsing import ParsingNone
from pgai.vectorizer.processing import ProcessingDefault

from ..destination import TableDestination
from . import register_migration


class LangChainCharacterTextSplitter_0_9(BaseModel, Chunker):
    implementation: Literal["character_text_splitter"]
    separator: str
    chunk_size: int
    chunk_column: str
    chunk_overlap: int
    is_separator_regex: bool

    @override
    def into_chunks(self, row: dict[str, Any], payload: str) -> list[str]:
        return []  # noop


class LangChainRecursiveCharacterTextSplitter_0_9(BaseModel, Chunker):
    implementation: Literal["recursive_character_text_splitter"]
    separators: list[str]
    chunk_size: int
    chunk_column: str
    chunk_overlap: int
    is_separator_regex: bool

    @override
    def into_chunks(self, row: dict[str, Any], payload: str) -> list[str]:
        return []  # noop


class Config_0_9(BaseModel):
    model_config = ConfigDict(extra="allow")
    version: str
    embedding: OpenAI | Ollama | VoyageAI | LiteLLM
    processing: ProcessingDefault
    chunking: (
        LangChainCharacterTextSplitter_0_9 | LangChainRecursiveCharacterTextSplitter_0_9
    ) = Field(..., discriminator="implementation")
    formatting: PythonTemplate | ChunkValue = Field(..., discriminator="implementation")


class Vectorizer_0_9(BaseModel):
    model_config = ConfigDict(extra="allow")
    config: Config_0_9
    target_table: str
    target_schema: str


@register_migration(
    "0.10.0",
    Vectorizer_0_9,
    "Migrate from no loading config to column_config, add "
    "parsing none, use proper chunking config",
)
def migrate_to_0_10_0(old_vectorizer: Vectorizer_0_9) -> dict[str, Any]:
    # use the data as is from the previous version so we modify whatever is needed
    result = old_vectorizer.model_dump()
    old_config = old_vectorizer.config

    result["config"]["loading"] = ColumnLoading(
        implementation="column", column_name=old_config.chunking.chunk_column
    )
    result["config"]["parsing"] = ParsingNone(implementation="none")
    match old_vectorizer.config.chunking.implementation:
        case "character_text_splitter":
            result["config"]["chunking"] = LangChainCharacterTextSplitter(
                implementation="character_text_splitter",
                separator=old_config.chunking.separator,  # type: ignore[reportUnknownVariableType]
                chunk_size=old_config.chunking.chunk_size,
                chunk_overlap=old_config.chunking.chunk_overlap,
                is_separator_regex=old_config.chunking.is_separator_regex,
            )
        case "recursive_character_text_splitter":
            result["config"]["chunking"] = LangChainRecursiveCharacterTextSplitter(
                implementation="recursive_character_text_splitter",
                separators=old_config.chunking.separators,  # type: ignore[reportUnknownVariableType]
                chunk_size=old_config.chunking.chunk_size,
                chunk_overlap=old_config.chunking.chunk_overlap,
                is_separator_regex=old_config.chunking.is_separator_regex,
            )
        case _:
            raise ValueError(
                f"Unknown chunking implementation: {old_config.chunking.implementation}"
            )

    result["config"]["destination"] = TableDestination(
        implementation="table",
        target_schema=old_vectorizer.target_schema,
        target_table=old_vectorizer.target_table,
    )

    return result



import subprocess
from pathlib import Path

import psycopg

import pgai
import pgai.semantic_catalog as sc
from pgai.semantic_catalog.vectorizer import OllamaConfig, OpenAIConfig
from tests.semantic_catalog.utils import PostgresContainer


async def setup(dump_url: str) -> None:
    pgai.install(dump_url)
    async with await psycopg.AsyncConnection.connect(dump_url) as con:
        catalog = await sc.create(con)
        assert catalog.id == 1, "catalog.id is not 1"
        assert catalog.name == "default", "catalog.name is not 'default'"
        # add a second embedding config
        await catalog.add_embedding(
            con,
            OpenAIConfig.create(
                model="text-embedding-3-small",
                dimensions=1536,
            ),
        )
        # add a third embedding config
        await catalog.add_embedding(
            con,
            OllamaConfig.create(
                model="nomic-embed-text",
                dimensions=768,
            ),
        )
        async with con.cursor() as cur:
            script = Path(__file__).parent.joinpath("vector_dump.sql").read_text()
            await cur.execute(script)  # pyright: ignore [reportArgumentType]
            await cur.execute("select count(*) from ai.semantic_catalog_obj_1")
            row = await cur.fetchone()
            assert row is not None, "no result for count of semantic_catalog_obj_1"
            assert row[0] > 0
            await cur.execute("select ai.sc_grant_admin('ada')")
            await cur.execute("select ai.sc_grant_write('default', 'vera')")
            await cur.execute("select ai.sc_grant_read('default', 'edith')")


async def test_dump_restore(container: PostgresContainer):
    container.drop_database("dump")
    container.drop_database("restore")

    container.drop_user("ada")
    container.drop_user("vera")
    container.drop_user("edith")
    container.drop_user("bot")

    container.create_database("restore")
    container.create_database("dump")

    container.create_user("ada")
    container.create_user("vera")
    container.create_user("edith")
    container.create_user("bot")

    dump_url = container.connection_string(database="dump")
    restore_url = container.connection_string(database="restore")
    await setup(dump_url)

    dump_expected = Path(__file__).parent.joinpath("dump.expected")
    dump_actual = Path(__file__).parent.joinpath("dump.actual")
    privs_expected = Path(__file__).parent.joinpath("privs.expected").read_text()

    dump_expected.unlink(missing_ok=True)
    dump_actual.unlink(missing_ok=True)

    # dump the dump db
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/dump_restore" {container.name}',
                'pg_dump -d "postgres://postgres@localhost:5432/dump"',
                "-F p --schema=postgres_air --schema=ai",
                "-f dump.expected",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert cp.returncode == 0, f"dump of dump db failed: {cp.stderr}"

    assert dump_expected.exists(), "dump.expected does not exist"

    # get the privileges in the dump db
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/dump_restore" {container.name}',
                'psql -d "postgres://postgres@localhost:5432/dump"',
                "-X",
                "-P format=aligned",
                "-f privs.sql",
                "-o privs_dump.actual",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert cp.returncode == 0, f"dump of dump db privs failed: {cp.stderr}"
    privs_actual = Path(__file__).parent.joinpath("privs_dump.actual").read_text()
    assert privs_actual == privs_expected

    # install pgvector in restore db
    async with (
        await psycopg.AsyncConnection.connect(restore_url) as con,
        con.cursor() as cur,
    ):
        await cur.execute("create extension vector")

    # restore to the restore db
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/dump_restore" {container.name}',
                'psql -d "postgres://postgres@localhost:5432/restore"',
                "-v ON_ERROR_STOP=1",
                "-f dump.expected",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert cp.returncode == 0, f"restore to the restore db failed: {cp.stderr}"

    # dump the restore db
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/dump_restore" {container.name}',
                'pg_dump -d "postgres://postgres@localhost:5432/restore"',
                "-F p --schema=postgres_air --schema=ai",
                "-f dump.actual",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert cp.returncode == 0, f"dump of the restore db failed: {cp.stderr}"

    # compare dump files of restore and dump databases
    expected = dump_expected.read_text()
    actual = dump_actual.read_text()
    assert expected == actual

    # get the privileges in the restore db
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/dump_restore" {container.name}',
                'psql -d "postgres://postgres@localhost:5432/restore"',
                "-X",
                "-P format=aligned",
                "-f privs.sql",
                "-o privs_restore.actual",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert cp.returncode == 0, f"dump of restore db privs failed: {cp.stderr}"
    privs_actual = Path(__file__).parent.joinpath("privs_restore.actual").read_text()
    assert privs_actual == privs_expected

    # load the postgres air schema into the restored database
    # now, the oids of the postgres air objects won't match the semantic catalog
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/data" {container.name}',
                '/usr/bin/psql -d "postgres://postgres@localhost:5432/restore"',
                "-v ON_ERROR_STOP=1",
                "-f postgres_air.sql",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert (
        cp.returncode == 0
    ), f"failed to load postgres_air into restore db: {cp.stderr}"

    async with (
        await psycopg.AsyncConnection.connect(restore_url, autocommit=True) as con,
        con.cursor() as cur,
    ):
        catalog = await sc.from_id(con, 1)

        # do we have all the object descriptions we expect?
        await cur.execute(f"""\
            select count(*)
            from ai.semantic_catalog_obj_{catalog.id}
        """)  # pyright: ignore [reportArgumentType]
        row = await cur.fetchone()
        assert row and row[0] == 16

        # fix the ids
        await catalog.fix_ids(con, con)

        # postgres_air.flight_summary should have been deleted
        await cur.execute(f"""\
            select count(*)
            from ai.semantic_catalog_obj_{catalog.id}
        """)  # pyright: ignore [reportArgumentType]
        row = await cur.fetchone()
        assert row and row[0] == 11

        # change the objtype to 'bob' across the board
        await cur.execute(f"""\
            update ai.semantic_catalog_obj_{catalog.id}
            set objtype = 'bob'
        """)  # pyright: ignore [reportArgumentType]

        # fix the names
        await catalog.fix_names(con, con)

        # objtype should have been fixed
        await cur.execute(f"""\
            select count(*)
            from ai.semantic_catalog_obj_{catalog.id}
            where objtype != 'bob'
        """)  # pyright: ignore [reportArgumentType]
        row = await cur.fetchone()
        assert row and row[0] == 11

        # swap the names of two tables
        async with con.transaction() as _:
            await cur.execute("""\
                create table postgres_air.bob();
                select ai.sc_set_table_desc
                ( 'postgres_air.bob'::regclass
                , 'this table has no columns'
                );
                alter table postgres_air.airport rename to fred;
                alter table postgres_air.bob rename to airport;
                alter table postgres_air.fred rename to bob;
            """)

        # fix the names
        await catalog.fix_names(con, con)

        # objnames should have been fixed
        await cur.execute(f"""\
            select description
            from ai.semantic_catalog_obj_{catalog.id}
            where objnames = array['postgres_air', 'airport']
        """)  # pyright: ignore [reportArgumentType]
        row = await cur.fetchone()
        assert row and row[0] == "this table has no columns"

        await cur.execute("""\
            select ai.sc_grant_read('default', 'bot');
            select ai.sc_grant_obj_read('default', 'bot');
        """)

    # get the bot's privileges to the objects
    cp = subprocess.run(
        " ".join(
            [
                f'docker exec -w "/tmp/tests/dump_restore" {container.name}',
                'psql -d "postgres://postgres@localhost:5432/restore"',
                "-X",
                "-P format=aligned",
                "-f obj_privs.sql",
                "-o obj_privs.actual",
            ]
        ),
        text=True,
        shell=True,
        capture_output=True,
    )
    assert cp.returncode == 0, f"dump of obj privs failed: {cp.stderr}"
    obj_actual = Path(__file__).parent.joinpath("obj_privs.actual").read_text()
    obj_expected = Path(__file__).parent.joinpath("obj_privs.expected").read_text()
    assert obj_actual == obj_expected



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_formatting_python_template():
    tests = [
        (
            """
            select ai.formatting_python_template()
            """,
            {
                "implementation": "python_template",
                "config_type": "formatting",
                "template": "$chunk",
            },
        ),
        (
            """
            select ai.formatting_python_template
            ( 'size: $size shape: $shape $chunk'
            )
            """,
            {
                "implementation": "python_template",
                "config_type": "formatting",
                "template": "size: $size shape: $shape $chunk",
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_validate_formatting():
    ok = [
        """
        select ai._validate_formatting
        ( ai.formatting_python_template()
        , 'public', 'thing'
        )
        """,
        """
        select ai._validate_formatting
        ( ai.formatting_python_template('color: $color weight: $weight $chunk')
        , 'public', 'thing'
        )
        """,
    ]
    bad = [
        (
            """
            select ai._validate_formatting
            ( ai.scheduling_none()
            , 'public', 'thing'
            )
            """,
            "invalid config_type for formatting config",
        ),
        (
            """
            select ai._validate_formatting
            ( '{"implementation": "jinja2", "config_type": "formatting"}'::jsonb
            , 'public', 'thing'
            )
            """,
            "unrecognized formatting implementation",
        ),
        (
            """
            select ai._validate_formatting
            ( ai.formatting_python_template
              ( 'color: $color weight: $weight height: $height' -- no $chunk
              )
            , 'public', 'thing'
            )
            """,
            "template must contain $chunk placeholder",
        ),
        (
            """
            select ai._validate_formatting
            ( ai.formatting_python_template
              ( 'color: $color weight: $weight height: $height $chunk'
              )
            , 'public', 'thing2' -- has a column named "chunk"
            )
            """,
            'formatting_python_template may not be used when source table has a column named "chunk"',
        ),
        (
            """
            select ai._validate_formatting
            ( ai.scheduling_none()
            , 'public', 'thing2' -- has a column named "chunk"
            )
            """,
            "invalid config_type for formatting config",
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            cur.execute("drop table if exists public.thing;")
            cur.execute("create table public.thing (id int, color text, weight float)")
            cur.execute("drop table if exists public.thing2;")
            cur.execute(
                "create table public.thing2 (id int, color text, weight float, chunk text)"
            )
            for query in ok:
                cur.execute(query)
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")



import os
from pathlib import Path

import psycopg
import pytest
from docker.types import Mount
from dotenv import load_dotenv

from .utils import PostgresContainer

load_dotenv()


def load_postgres_air(container: PostgresContainer):
    script1 = Path(__file__).parent.joinpath("data", "postgres_air.sql").read_text()
    script2 = (
        Path(__file__).parent.joinpath("data", "postgres_air_extra.sql").read_text()
    )
    with (
        psycopg.connect(container.connection_string(database="postgres_air")) as con,
        con.transaction() as _,
        con.cursor() as cur,
    ):
        cur.execute(script1)  # pyright: ignore [reportArgumentType]
        cur.execute(script2)  # pyright: ignore [reportArgumentType]
    script3 = Path(__file__).parent.joinpath("data", "cagg.sql").read_text()
    with (
        psycopg.connect(
            container.connection_string(database="postgres_air"),
            autocommit=True,
        ) as con,
        con.cursor() as cur,
    ):
        for query in script3.split(";"):
            if query.strip():
                cur.execute(query)  # pyright: ignore [reportArgumentType]


@pytest.fixture(scope="session")
def container():
    dont_kill_container = os.getenv("DONT_KILL_CONTAINER") is not None
    container = PostgresContainer.get_or_create(
        mounts=[
            Mount(
                type="bind",
                source=str(Path(__file__).parent.resolve()),
                target="/tmp/tests",
            )
        ]
    )
    container.drop_database("postgres_air")
    container.create_database("postgres_air")
    load_postgres_air(container)
    yield container
    if not dont_kill_container:
        container.container.stop()
        container.container.remove()



import os

import psycopg
import pytest
from psycopg.sql import SQL

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_loading_column():
    tests = [
        (
            "select ai.loading_column('content')",
            {
                "config_type": "loading",
                "implementation": "column",
                "column_name": "content",
                "retries": 6,
            },
        ),
        (
            "select ai.loading_column('content', 10)",
            {
                "config_type": "loading",
                "implementation": "column",
                "column_name": "content",
                "retries": 10,
            },
        ),
    ]

    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_loading_uri():
    tests = [
        (
            "select ai.loading_uri('s3_uri')",
            {
                "config_type": "loading",
                "implementation": "uri",
                "column_name": "s3_uri",
                "retries": 6,
            },
        ),
        (
            "select ai.loading_uri('s3_uri', 3)",
            {
                "config_type": "loading",
                "implementation": "uri",
                "column_name": "s3_uri",
                "retries": 3,
            },
        ),
        (
            "select ai.loading_uri('s3_uri', aws_role_arn => 'arn:aws:iam::account:role/role-name-with-path')",
            {
                "config_type": "loading",
                "implementation": "uri",
                "column_name": "s3_uri",
                "retries": 6,
                "aws_role_arn": "arn:aws:iam::account:role/role-name-with-path",
            },
        ),
    ]

    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_validate_loading():
    ok = [
        # loading raw text data from a text column
        SQL("""
        select ai._validate_loading
        ( ai.loading_column('body'), 'public', 'thing' )
        """),
        # loading a document from a bytea column
        SQL("""
        select ai._validate_loading
        ( ai.loading_column('document'), 'public', 'thing' )
        """),
        # loading a document from a uri stored in any text column
        # (we do not validate the uri, smart-open does on runtime)
        SQL("""
        select ai._validate_loading
        ( ai.loading_uri('body'), 'public', 'thing' )
        """),
        # setting the aws_role_arn parameter
        SQL("""
        select ai._validate_loading
        ( ai.loading_uri('body', aws_role_arn => 'arn:aws:iam::account:role/role-name-with-path'), 'public', 'thing' )
        """),
    ]
    bad = [
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_column(), 'public', 'thing' )
            """),
            "function ai.loading_column() does not exist",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_uri(), 'public', 'thing' )
            """),
            "function ai.loading_uri() does not exist",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_column('column_does_not_exist'), 'public', 'thing' )
            """),
            "column_name in config does not exist in the table: column_does_not_exist",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_column('weight'), 'public', 'thing' )
            """),
            "column_name weight in config is of invalid type float8. Supported types are: text, varchar, char, bpchar, bytea",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_uri('weight'), 'public', 'thing' )
            """),
            "column_name weight in config is of invalid type float8. Supported types are: text, varchar, char, bpchar, bytea",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_column('non_existing_column'), 'public', 'thing' )
            """),
            "column_name in config does not exist in the table: non_existing_column",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_uri('non_existing_column'), 'public', 'thing' )
            """),
            "column_name in config does not exist in the table: non_existing_column",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_uri('document'), 'public', 'thing' )
            """),
            "the type of the column `document` in config is not compatible with `uri` loading implementation (type should be either text, varchar, char, bpchar, or bytea)",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.scheduling_none(), 'public', 'thing' )
            """),
            "invalid config_type for loading config",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_column('body', -1), 'public', 'thing' )
            """),
            "invalid loading config, retries must be a non-negative integer",
        ),
        (
            SQL("""
            select ai._validate_loading
            ( ai.loading_uri('body', aws_role_arn => 'foo_bar'), 'public', 'thing' )
            """),
            "invalid loading config, aws_role_arn must match arn:aws:iam::*:role/*",
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            cur.execute("drop table if exists public.thing;")
            cur.execute(
                "create table public.thing (id int, color text, weight float, body text,document bytea)"
            )
            for query in ok:
                cur.execute(query)
                assert True
            for query, err in bad:
                with pytest.raises(psycopg.ProgrammingError) as ex:
                    cur.execute(query)
                assert str(ex.value).startswith(err)



import json
import time
from functools import cached_property
from pathlib import Path
from typing import Any

import docker
import psycopg
from docker import DockerClient
from docker.models.containers import Container
from docker.models.images import Image
from docker.types import Mount
from psycopg.sql import SQL, Identifier

from pgai.semantic_catalog import describe, loader
from pgai.semantic_catalog.models import Procedure, Table, View

CONTAINER_NAME = "pgai-semantic-catalog"


class PostgresContainer:
    def __init__(self, container: Container):
        self._container = container

    @property
    def container(self) -> Container:
        return self._container

    @property
    def name(self) -> str | None:
        return self.container.name

    @cached_property
    def host(self) -> str:
        return self._container.ports["5432/tcp"][0]["HostIp"]

    @cached_property
    def port(self) -> str:
        return self._container.ports["5432/tcp"][0]["HostPort"]

    def connection_string(
        self, user: str = "postgres", database: str = "postgres"
    ) -> str:
        return f"postgres://{user}@{self.host}:{self.port}/{database}"

    def __str__(self):  # pyright: ignore [reportImplicitOverride]
        return self.connection_string()

    def wait_for(self):
        trigger = "database system is ready to accept connections"
        count = 0
        for line in self._container.logs(stream=True, follow=True):
            if trigger in line.decode("utf-8").strip():
                count = count + 1
                if count >= 2:
                    break

    def is_ready(
        self, timeout_seconds: float = 30.0, interval_seconds: float = 0.2
    ) -> bool:
        url = self.connection_string()
        start = time.time()
        while time.time() - start < timeout_seconds:
            try:
                with (
                    psycopg.connect(url, connect_timeout=1) as con,
                    con.cursor() as cur,
                ):
                    cur.execute("select 1")
                    cur.fetchall()
                return True
            except psycopg.OperationalError as _:
                time.sleep(interval_seconds)
        return False

    def drop_database(self, database_name: str, force: bool = True) -> None:
        with (
            psycopg.connect(self.connection_string(), autocommit=True) as con,
            con.cursor() as cur,
        ):
            query = SQL("drop database if exists {}").format(Identifier(database_name))
            if force:
                query = query + SQL(" with (force)")
            cur.execute(query)

    def create_database(self, database_name: str, owner: str | None = None) -> None:
        with (
            psycopg.connect(self.connection_string(), autocommit=True) as con,
            con.cursor() as cur,
        ):
            query = SQL("create database {}").format(Identifier(database_name))
            if owner is not None:
                query = query + SQL("owner={}").format(Identifier(owner))
            cur.execute(query)

    def drop_user(self, user_name: str) -> None:
        with psycopg.connect(self.connection_string()) as con, con.cursor() as cur:
            query = SQL("drop user if exists {}").format(Identifier(user_name))
            cur.execute(query)

    def create_user(self, user_name: str) -> None:
        with psycopg.connect(self.connection_string()) as con, con.cursor() as cur:
            query = SQL("create user {}").format(Identifier(user_name))
            cur.execute(query)

    @staticmethod
    def get(container_name: str = CONTAINER_NAME) -> "PostgresContainer | None":
        client: DockerClient = docker.from_env()
        containers: list[Container] = client.containers.list(  # pyright: ignore [reportUnknownVariableType, reportUnknownMemberType]
            filters={"name": container_name},
            all=True,
        )
        if not containers:
            return None
        return PostgresContainer(container=containers[0])  # pyright: ignore [reportUnknownVariableType, reportUnknownArgumentType]

    @staticmethod
    def get_or_create(
        container_name: str = CONTAINER_NAME,
        port: int = 5678,
        mounts: list[Mount] | None = None,
    ) -> "PostgresContainer":
        client: DockerClient = docker.from_env()
        maybe: PostgresContainer | None = PostgresContainer.get(container_name)
        if maybe is None:
            dockerfile = Path(__file__).parent.parent.parent.parent.joinpath(
                "extension"
            )
            image: Image = client.images.build(
                path=str(dockerfile), target="pgai-test-db"
            )[0]
            container = PostgresContainer(  # pyright: ignore [reportCallIssue, reportUnknownArgumentType]
                container=client.containers.run(  # pyright: ignore [reportCallIssue, reportUnknownArgumentType]
                    image=image.id,  # pyright: ignore [reportArgumentType]
                    name=container_name,
                    detach=True,
                    ports={"5432/tcp": ("127.0.0.1", port)},
                    environment={
                        "POSTGRES_HOST_AUTH_METHOD": "trust",
                        "POSTGRES_INITDB_ARGS": " ".join(
                            [
                                "-c shared_preload_libraries='timescaledb, pgextwlist'",
                                "-c extwlist.extensions='ai,vector'",
                            ]
                        ),
                    },
                    mounts=mounts,
                )
            )
            container.wait_for()
        else:
            container: PostgresContainer = maybe
            if container.container.status != "running":
                container.container.start()
                container.wait_for()
        assert container is not None
        while "5432/tcp" not in container.container.ports:
            time.sleep(0.1)
            maybe = PostgresContainer.get(container_name)
            if maybe is None:
                raise RuntimeError(f"Could not find container {container_name}")
            container = maybe
        assert container is not None
        if not container.is_ready():
            raise RuntimeError("Postgres container not ready.")
        return container


async def gen_tables_json(container: PostgresContainer):
    stuff = {}
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        oids = await describe.find_tables(con)
        tables = await loader.load_tables(con, oids)
        for table in tables:
            stuff[table.table_name] = table.model_dump()
    Path(__file__).parent.joinpath("data", "tables.json").write_text(
        json.dumps(stuff, indent=2)
    )


async def gen_views_json(container: PostgresContainer):
    stuff = {}
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        oids = await describe.find_views(con)
        views = await loader.load_views(con, oids)
        for view in views:
            stuff[view.view_name] = view.model_dump()
    Path(__file__).parent.joinpath("data", "views.json").write_text(
        json.dumps(stuff, indent=2)
    )


async def gen_procs_json(container: PostgresContainer):
    stuff = {}
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        oids = await describe.find_procedures(con)
        procs = await loader.load_procedures(con, oids)
        for proc in procs:
            stuff[proc.proc_name] = proc.model_dump()
    Path(__file__).parent.joinpath("data", "procedures.json").write_text(
        json.dumps(stuff, indent=2)
    )


def get_table_dict() -> dict[str, Table]:
    raw: dict[str, Any] = json.loads(
        Path(__file__).parent.joinpath("data", "tables.json").read_text()
    )
    return {k: Table(**v) for k, v in raw.items()}


def get_tables() -> list[Table]:
    d = get_table_dict()
    keys = [k for k in d]
    keys.sort()
    vals: list[Table] = []
    for k in keys:
        val = d[k]
        vals.append(val)
    return vals


def get_view_dict() -> dict[str, View]:
    raw: dict[str, Any] = json.loads(
        Path(__file__).parent.joinpath("data", "views.json").read_text()
    )
    return {k: View(**v) for k, v in raw.items()}


def get_views() -> list[View]:
    d = get_view_dict()
    keys = [k for k in d]
    keys.sort()
    vals: list[View] = []
    for k in keys:
        val = d[k]
        vals.append(val)
    return vals


def get_procedure_dict() -> dict[str, Procedure]:
    raw: dict[str, Any] = json.loads(
        Path(__file__).parent.joinpath("data", "procedures.json").read_text()
    )
    return {k: Procedure(**v) for k, v in raw.items()}


def get_procedures() -> list[Procedure]:
    d = get_procedure_dict()
    keys = [k for k in d]
    keys.sort()
    vals: list[Procedure] = []
    for k in keys:
        val = d[k]
        vals.append(val)
    return vals


async def load_airports(con: psycopg.AsyncConnection) -> None:
    airport_data = Path(__file__).parent.joinpath("data", "airport.sql").read_text()
    async with (
        con.cursor() as cur,
        cur.copy("copy postgres_air.airport from stdin with (format csv)") as cpy,
    ):
        await cpy.write(airport_data)



from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row

from tests.vectorizer.cli.conftest import (
    TestDatabase,
    run_vectorizer_worker,
    setup_source_table,
)


def configure_same_table_vectorizer(
    connection: Connection,
    table_name: str,
) -> int:
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute(f"""
            SELECT ai.create_vectorizer(
                '{table_name}'::regclass,
                destination => ai.destination_column('embedding'),
                embedding => ai.embedding_openai('text-embedding-ada-002', 1536),
                chunking => ai.chunking_none(),
                loading => ai.loading_column('content')
            )
        """)  # type: ignore

        vectorizer_id: int = int(cur.fetchone()["create_vectorizer"])  # type: ignore
        return vectorizer_id


def test_same_table_vectorizer(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """
    Test that destination_column works and adds the embeddings to the original table
    """
    _, con = cli_db
    table_name = setup_source_table(con, 2)

    # Given a vectorizer that has 2 items and a batch size of 1
    vectorizer_id = configure_same_table_vectorizer(con, table_name)

    with vcr_.use_cassette("same_table_vectorizer.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

        assert result.exit_code == 0
        assert "finished processing vectorizer" in result.output.lower()

        with con.cursor(row_factory=dict_row) as cur:
            cur.execute("SELECT * FROM blog;")
            results = cur.fetchone()
            assert results is not None
            assert results["embedding"] is not None

        with con.cursor(row_factory=dict_row) as cur:
            cur.execute("SELECT id, embedding FROM blog order by id;")
            results_pre_update = cur.fetchall()

            cur.execute("UPDATE blog set content = 'new content' where id = 1;")
            con.commit()
            result = run_vectorizer_worker(cli_db_url, vectorizer_id)
            assert result.exit_code == 0

            cur.execute("SELECT id, embedding FROM blog order by id;")
            results_post_update = cur.fetchall()

            assert (
                results_post_update[0]["embedding"]
                != results_pre_update[0]["embedding"]
            )
            assert (
                results_post_update[1]["embedding"]
                == results_pre_update[1]["embedding"]
            )

            cur.execute("SELECT * FROM ai.vectorizer_status;")
            row = cur.fetchone()
            assert row is not None
            assert row["source_table"] == "public" + "." + table_name
            assert row["target_table"] is None
            assert row["view"] is None
            assert row["embedding_column"] == "embedding"
            assert row["pending_items"] == 0
            assert not row["disabled"]

            # Check deletes just work
            cur.execute("DELETE FROM blog where id = 1;")
            con.commit()
            result = run_vectorizer_worker(cli_db_url, vectorizer_id)
            assert result.exit_code == 0



from pathlib import Path

from jinja2 import Environment

from pgai.vectorizer.generate.function_parser import (
    PostgresFunction,
    VectorizerParameter,
)

TEMPLATE = '''"""This file is generated by the configuration generator
in generate/config_generator.py. Do not modify by hand."""
from dataclasses import dataclass
from typing import ClassVar, Any, TypeAlias
from datetime import datetime, timedelta

JsonDict: TypeAlias = dict[str, Any]

def format_sql_params(params: dict[str, str | None | bool | list[str]]) -> str:
    """Format dictionary of parameters into SQL argument string without any quoting."""
    formatted: list[str] = []
    for key, value in params.items():
        if value is None:
            continue
        elif isinstance(value, bool):
            formatted.append(f"{key}=>{str(value).lower()}")
        elif isinstance(value, list):
            array_list = ",".join(f"E'{v}'" for v in value)
            formatted.append(f"{key}=>ARRAY[{array_list}]")
        elif isinstance(value, timedelta):
            formatted.append(f"{key}=>'{value.seconds} seconds'")
        else:
            formatted.append(f"{key}=> '{value}'")
    return ", ".join(formatted)

class SQLArgumentMixin:
    arg_type: ClassVar[str]
    function_name: ClassVar[str]

    def to_sql_argument(self) -> str:
        params = {}
        for field_name in self.__dataclass_fields__:  # type: ignore
            if field_name != "arg_type" and field_name != "function_name":
                value = getattr(self, field_name)  # type: ignore
                if value is not None:
                    params[field_name] = value

        return f", {self.arg_type} => {self.function_name}({format_sql_params(params)})"  # type: ignore

{% for function in functions %}


@dataclass
class {{ function.name|replace('ai.', '')|replace('_', ' ')|title|replace(' ', '') }}Config(SQLArgumentMixin):
    """Configuration for {{ function.name }} function."""
    arg_type: ClassVar[str] = "{{ function.name|split('.')|last|split('_')|first }}"
    function_name: ClassVar[str] = "{{ function.name }}"

    {% for param in function.parameters %}
    {{ param.name }}: {{ param.python_type }}{% if param.is_required == False %} = None{% endif %}

    {% endfor %}
{% endfor %}
'''

VECTORIZER_TEMPLATE = '''"""This file is generated by the configuration generator
in generate/config_generator.py. Do not modify by hand."""
from typing import Any
from dataclasses import dataclass

from .configuration import SQLArgumentMixin

# Dynamic imports based on parameter types
from .configuration import (
    {% for function in functions %}    {{ function.name|replace('ai.', '')|replace('_', ' ')|title|replace(' ', '') }}Config{%- if not loop.last %},
    {% endif %}{% endfor %}

)


@dataclass
class CreateVectorizer:
    """Parameters for ai.create_vectorizer function"""
    {% for param in parameters %}
    {{ param.name }}: {{ param.python_type }}{% if not param.is_required %} = None{% endif %}
    
    {% endfor %}

    def to_sql(self) -> str:
        """Convert parameters to SQL function call"""
        parts = ["SELECT ai.create_vectorizer("]

        parts.append(f"'{self.source}'::regclass")

        # Handle all config objects that implement SQLArgumentMixin
        for field_name, field_value in self.__dict__.items():
            if isinstance(field_value, SQLArgumentMixin):
                parts.append(field_value.to_sql_argument())
            elif field_value is not None and field_name not in ('source',):
                # Handle primitive types
                if isinstance(field_value, bool):
                    parts.append(f", {field_name} => {str(field_value).lower()}")
                elif isinstance(field_value, list):
                    values = ",".join(f"'{v}'" for v in field_value)  # type: ignore
                    parts.append(f", {field_name} => ARRAY[{values}]")
                else:
                    parts.append(f", {field_name} => '{field_value}'")

        parts.append(")")
        return "\\n".join(parts)
'''


def generate_vectorizer_params(
    parameters: list[VectorizerParameter],
    output_file: Path,
    functions: list[PostgresFunction],
) -> None:
    """Generate CreateVectorizerParams class from metadata"""
    env = Environment(trim_blocks=True, lstrip_blocks=True)
    template = env.from_string(VECTORIZER_TEMPLATE)

    output = template.render(parameters=parameters, functions=functions)
    output_file.write_text(output)


def generate_config_classes(
    functions: list[PostgresFunction], output_file: Path
) -> None:
    """Generate Python configuration classes from function metadata."""
    env = Environment(trim_blocks=True, lstrip_blocks=True)

    # Add custom filters
    env.filters["split"] = lambda x, sep: x.split(sep)  # type: ignore

    template = env.from_string(TEMPLATE)
    output = template.render(functions=functions)
    output_file.write_text(output)



"""
The migration template for these tests lies in
fixtures/migrations/generic_vectorizer.py.template
Each test generates a different migration, runs it, and verifies the configuration
This doesn't test if the vectorizer behaves as expected,
but it validates that the create_vectorizer call works.

If you add new configuration options, add a test for them here.
Either in an existing test case or a new one.
The config classes for migrations lie in pgai/alembic/configuration.py
If the migration and the validation definitions overlap,
it is possible to define a parent class in base.py
see e.g. CharacterTextSplitterConfig as an example
"""

import textwrap
from datetime import timedelta
from pathlib import Path
from typing import Any

from alembic.command import downgrade, upgrade
from alembic.config import Config
from sqlalchemy import Engine, text

from pgai.vectorizer.configuration import (
    ChunkingCharacterTextSplitterConfig,
    ChunkingRecursiveCharacterTextSplitterConfig,
    EmbeddingOllamaConfig,
    EmbeddingOpenaiConfig,
    EmbeddingVoyageaiConfig,
    FormattingPythonTemplateConfig,
    IndexingHnswConfig,
    LoadingColumnConfig,
    ProcessingDefaultConfig,
    SchedulingTimescaledbConfig,
)
from tests.vectorizer.extensions.conftest import load_template


def create_vectorizer_migration(migrations_dir: Path, vectorizer_config: str) -> None:
    """Create migration script with the given vectorizer configuration"""
    versions_dir = migrations_dir / "versions"
    migration_content = load_template(
        "migrations/generic_vectorizer.py.template",
        revision_id="001",
        revises="",
        create_date="2024-03-19 10:00:00.000000",
        down_revision="None",
        vectorizer_config=vectorizer_config,
    )

    with open(versions_dir / "001_generic_vectorizer.py", "w") as f:
        f.write(migration_content)


def create_vectorizer_config_code(**kwargs: Any) -> str:
    """Convert configuration objects to valid Python code for the migration template"""
    return textwrap.dedent(f"""
        op.create_vectorizer(
            source='public.blog',
            **{kwargs}
        )
    """).strip()


def test_openai_vectorizer(
    alembic_config: Config,
    initialized_engine: Engine,
):
    """Test OpenAI vectorizer configuration"""
    config = create_vectorizer_config_code(
        loading=LoadingColumnConfig("content"),
        embedding=EmbeddingOpenaiConfig(
            model="text-embedding-3-small",
            dimensions=768,
            chat_user="test_user",
            api_key_name="TEST_OPENAI_KEY",
        ),
        chunking=ChunkingCharacterTextSplitterConfig(
            chunk_size=256,
            chunk_overlap=20,
            separator="\n",
            is_separator_regex=False,
        ),
    )

    migrations_dir = Path(alembic_config.get_main_option("script_location"))  # type: ignore
    create_vectorizer_migration(migrations_dir, config)

    # Run upgrade
    upgrade(alembic_config, "head")

    # Verify configuration
    with initialized_engine.connect() as conn:
        vectorizer = conn.execute(text("SELECT * FROM ai.vectorizer")).fetchone()
        assert vectorizer is not None
        config = str(dict(vectorizer._mapping))  # type: ignore
        assert "text-embedding-3-small" in config
        assert "768" in config
        assert "test_user" in config

    # Run downgrade
    downgrade(alembic_config, "base")


def test_ollama_vectorizer(
    alembic_config: Config,
    initialized_engine: Engine,
):
    """Test Ollama vectorizer configuration"""
    config = create_vectorizer_config_code(
        loading=LoadingColumnConfig("content"),
        embedding=EmbeddingOllamaConfig(
            model="nomic-embed-text",
            dimensions=768,
            base_url="http://localhost:11434",
            keep_alive="5m",
        ),
        chunking=ChunkingRecursiveCharacterTextSplitterConfig(
            chunk_size=300,
            chunk_overlap=30,
            separators=["\n\n", "\n", "; "],
            is_separator_regex=False,
        ),
        formatting=FormattingPythonTemplateConfig(
            template="Title: $title\nContent: $chunk"
        ),
    )

    migrations_dir = Path(alembic_config.get_main_option("script_location"))  # type: ignore
    create_vectorizer_migration(migrations_dir, config)

    # Run upgrade
    upgrade(alembic_config, "head")

    # Verify configuration
    with initialized_engine.connect() as conn:
        vectorizer = conn.execute(text("SELECT * FROM ai.vectorizer")).fetchone()
        assert vectorizer is not None
        config = str(dict(vectorizer._mapping))  # type: ignore
        assert "nomic-embed-text" in config
        assert "http://localhost:11434" in config

    # Run downgrade
    downgrade(alembic_config, "base")


def test_voyage_vectorizer(
    alembic_config: Config,
    initialized_engine: Engine,
):
    """Test VoyageAI vectorizer configuration"""

    # create the ai extension to test the timescaledb scheduling
    with initialized_engine.connect() as conn:
        conn.execute(text("CREATE EXTENSION IF NOT EXISTS ai CASCADE;"))
        conn.commit()

    config = create_vectorizer_config_code(
        loading=LoadingColumnConfig("content"),
        embedding=EmbeddingVoyageaiConfig(
            model="voyage-ai-1",
            dimensions=256,
        ),
        chunking=ChunkingRecursiveCharacterTextSplitterConfig(),
        indexing=IndexingHnswConfig(
            min_rows=10000,
            opclass="vector_l1_ops",
            m=32,
            ef_construction=128,
            create_when_queue_empty=True,
        ),
        scheduling=SchedulingTimescaledbConfig(
            schedule_interval=timedelta(minutes=10), timezone="UTC", fixed_schedule=True
        ),
        processing=ProcessingDefaultConfig(batch_size=100, concurrency=2),
    )

    migrations_dir = Path(alembic_config.get_main_option("script_location"))  # type: ignore
    create_vectorizer_migration(migrations_dir, config)

    # Run upgrade
    upgrade(alembic_config, "head")

    # Verify configuration
    with initialized_engine.connect() as conn:
        vectorizer = conn.execute(text("SELECT * FROM ai.vectorizer")).fetchone()
        assert vectorizer is not None
        config = str(dict(vectorizer._mapping))  # type: ignore
        assert "voyage-ai-1" in config
        assert "256" in config
        assert "document" in config
        assert "vector_l1_ops" in config
        assert "32" in config  # m parameter
        assert "128" in config  # ef_construction

    # Run downgrade
    downgrade(alembic_config, "base")


def test_hnsw_vectorizer(
    alembic_config: Config,
    initialized_engine: Engine,
):
    """Test HNSW vectorizer configuration"""

    # create the ai extension to test the timescaledb scheduling
    # (and that's needed for auto indexing)
    with initialized_engine.connect() as conn:
        conn.execute(text("CREATE EXTENSION IF NOT EXISTS ai CASCADE;"))
        conn.commit()

    config = create_vectorizer_config_code(
        loading=LoadingColumnConfig("content"),
        embedding=EmbeddingOpenaiConfig(
            model="text-embedding-3-small",
            dimensions=768,
            api_key_name="TEST_OPENAI_KEY",
        ),
        chunking=ChunkingCharacterTextSplitterConfig(
            chunk_size=200,
            chunk_overlap=25,
            separator=" ",
            is_separator_regex=False,
        ),
        indexing=IndexingHnswConfig(
            min_rows=50000,
            opclass="vector_l1_ops",
            m=16,
            ef_construction=64,
            create_when_queue_empty=True,
        ),
        scheduling=SchedulingTimescaledbConfig(
            schedule_interval=timedelta(minutes=10), fixed_schedule=False
        ),
    )

    migrations_dir = Path(alembic_config.get_main_option("script_location"))  # type: ignore
    create_vectorizer_migration(migrations_dir, config)

    # Run upgrade
    upgrade(alembic_config, "head")

    # Verify configuration
    with initialized_engine.connect() as conn:
        vectorizer = conn.execute(text("SELECT * FROM ai.vectorizer")).fetchone()
        assert vectorizer is not None
        config = str(dict(vectorizer._mapping))  # type: ignore
        assert "vector_l1_ops" in config
        assert "16" in config  # m parameter
        assert "64" in config  # ef_construction

    # Run downgrade
    downgrade(alembic_config, "base")



import os

import psycopg
import pytest
from psycopg.rows import namedtuple_row

from db.tests.conftest import detailed_notice_handler


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def setup_website_blog_table():
    with psycopg.connect(
        db_url("test"), autocommit=True, row_factory=namedtuple_row
    ) as con:
        con.add_notice_handler(detailed_notice_handler)
        with con.cursor() as cur:
            cur.execute("drop schema if exists website cascade")
            cur.execute("create schema website")
            cur.execute("drop table if exists website.blog")
            cur.execute("""
                create table website.blog
                ( id int not null generated always as identity
                , title text not null
                , published timestamptz
                , body text not null
                , primary key (title, published)
                )
            """)
            cur.execute("""
                insert into website.blog(title, published, body)
                values
                  ('how to cook a hot dog', '2024-01-06'::timestamptz, 'put it on a hot grill')
                , ('how to make a sandwich', '2023-01-06'::timestamptz, 'put a slice of meat between two pieces of bread')
                , ('how to make stir fry', '2022-01-06'::timestamptz, 'pick up the phone and order takeout')
            """)


@pytest.mark.skipif(
    os.getenv("PG_MAJOR") == "15", reason="extension does not support pg15"
)
def test_named_vectorizer_timescaledb():
    setup_website_blog_table()
    with psycopg.connect(
        db_url("test"), autocommit=True, row_factory=namedtuple_row
    ) as con:
        con.add_notice_handler(detailed_notice_handler)
        with con.cursor() as cur:
            # create a vectorizer for the blog table
            # language=PostgreSQL
            cur.execute("""
                select ai.create_vectorizer
                ( 'website.blog'::regclass
                , name => 'blog_vectorizer'
                , loading=>ai.loading_column('body')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
                , scheduling=>ai.scheduling_none()
                , destination=>ai.destination_column('embedding1')
                , chunking=>ai.chunking_none()
                );
                """)

            cur.execute("create extension ai cascade")
            cur.execute(
                "select set_config('ai.external_functions_executor_url', 'http://0.0.0.0:8000', false)"
            )
            cur.execute(
                "select set_config('ai.external_functions_executor_events_path', '/api/v1/events', false)"
            )
            cur.execute("select ai.execute_vectorizer('blog_vectorizer')")

            with pytest.raises(psycopg.errors.NoDataFound):
                cur.execute("select ai.execute_vectorizer('non_existent_vectorizer')")


def test_named_vectorizer():
    setup_website_blog_table()
    with psycopg.connect(
        db_url("test"), autocommit=True, row_factory=namedtuple_row
    ) as con:
        con.add_notice_handler(detailed_notice_handler)
        with con.cursor() as cur:
            # create a vectorizer for the blog table
            # language=PostgreSQL
            cur.execute("""
            select ai.create_vectorizer
            ( 'website.blog'::regclass
            , name => 'blog_vectorizer'
            , loading=>ai.loading_column('body')
            , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
            , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
            , scheduling=>ai.scheduling_none()
            , destination=>ai.destination_column('embedding1')
            , chunking=>ai.chunking_none()
            );
            """)

            cur.execute(
                "select ai.vectorizer_queue_pending(name => %s)", ("blog_vectorizer",)
            )
            pending = cur.fetchone()[0]
            assert pending == 3

            # testing ai.vectorizer_status view filtering by vectorizer name
            cur.execute(
                "select pending_items from ai.vectorizer_status where name = %s",
                ("blog_vectorizer",),
            )
            status_pending_items = cur.fetchone()[0]
            assert status_pending_items == 3

            # create another vectorizer with same name but if not exists set
            cur.execute("""
            select ai.create_vectorizer
            ( 'website.blog'::regclass
            , name => 'blog_vectorizer'
            , loading=>ai.loading_column('body')
            , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
            , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
            , scheduling=>ai.scheduling_none()
            , destination=>ai.destination_column('embedding1')
            , chunking=>ai.chunking_none()
            , if_not_exists => true
            );
            """)

            cur.execute(
                "select ai.vectorizer_queue_pending(name => %s)", ("blog_vectorizer",)
            )
            pending = cur.fetchone()[0]
            assert pending == 3

            with pytest.raises(psycopg.errors.DuplicateObject):
                # create a vectorizer with same name but if not exists set to false
                cur.execute("""
                select ai.create_vectorizer
                ( 'website.blog'::regclass
                , name => 'blog_vectorizer'
                , loading=>ai.loading_column('body')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
                , scheduling=>ai.scheduling_none()
                , destination=>ai.destination_column('embedding1')
                , chunking=>ai.chunking_none()
                , if_not_exists => false
                );
                """)

            # create a vectorizer with no name check default name
            cur.execute("""
               select ai.create_vectorizer
               ( 'website.blog'::regclass
               , loading=>ai.loading_column('body')
               , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
               , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
               , scheduling=>ai.scheduling_none()
               , destination=>ai.destination_column('embedding1')
               , chunking=>ai.chunking_none()
               , if_not_exists => false
               );
               """)

            vectorizer_id_2 = cur.fetchone()[0]
            cur.execute(
                """
                select name from ai.vectorizer where id = %s
            """,
                (vectorizer_id_2,),
            )
            vectorizer_name = cur.fetchone()[0]
            assert vectorizer_name == "website_blog_embedding1"

            # Test fetch errors by vectorizer name
            cur.execute(
                "insert into ai._vectorizer_errors (id, message) values (%s, %s)",
                (vectorizer_id_2, "test error message"),
            )

            cur.execute(
                "select * from ai.vectorizer_errors where name = %s",
                (vectorizer_name,),
            )

            error = cur.fetchone()
            assert error.message == "test error message"

            # create a vectorizer with no name check default name
            cur.execute("""
               select ai.create_vectorizer
               ( 'website.blog'::regclass
               , loading=>ai.loading_column('body')
               , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
               , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
               , scheduling=>ai.scheduling_none()
               , destination=>ai.destination_table()
               , chunking=>ai.chunking_none()
               , if_not_exists => false
               );
               """)
            vectorizer_id_3 = cur.fetchone()[0]
            cur.execute(
                """
                            select name from ai.vectorizer where id = %s
                        """,
                (vectorizer_id_3,),
            )
            vectorizer_name = cur.fetchone()[0]
            assert vectorizer_name == "website_blog_embedding_store"

            # try to recreate the vectorizer with the same default name
            # and if not exists set to true
            cur.execute("""
               select ai.create_vectorizer
               ( 'website.blog'::regclass
               , loading=>ai.loading_column('body')
               , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
               , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
               , scheduling=>ai.scheduling_none()
               , destination=>ai.destination_table()
               , chunking=>ai.chunking_none()
               , if_not_exists => true
               );
               """)
            vectorizer_id_3 = cur.fetchone()[0]
            cur.execute(
                """
                            select name from ai.vectorizer where id = %s
                        """,
                (vectorizer_id_3,),
            )
            vectorizer_name = cur.fetchone()[0]
            assert vectorizer_name == "website_blog_embedding_store"

            # make sure their is an error if if_not_exists is false
            with pytest.raises(psycopg.errors.DuplicateObject):
                cur.execute("""
                select ai.create_vectorizer
                ( 'website.blog'::regclass
                , loading=>ai.loading_column('body')
                , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                , formatting=>ai.formatting_python_template('title: $title published: $published $chunk')
                , scheduling=>ai.scheduling_none()
                , destination=>ai.destination_table()
                , chunking=>ai.chunking_none()
                , if_not_exists => false
                );
                """)

            # test functions with vectorizer name can be called without error
            cur.execute(
                "select ai.disable_vectorizer_schedule(%s)",
                ("website.blog_embedding_store",),
            )
            cur.execute(
                "select ai.enable_vectorizer_schedule(%s)",
                ("website.blog_embedding_store",),
            )
            cur.execute(
                "select ai.vectorizer_queue_pending(%s)",
                ("website.blog_embedding_store",),
            )
            cur.execute(
                "select ai.drop_vectorizer(%s, drop_all => true)",
                ("website.blog_embedding_store",),
            )



import shutil
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
from click.testing import CliRunner


def test_download_models():
    """Test that download-models command works correctly."""
    # Create a temporary directory for the test
    with patch("docling.utils.model_downloader.download_models") as mock_download:
        from pgai.cli import download_models
        from pgai.vectorizer.parsing import DOCLING_CACHE_DIR

        # Run the CLI command
        result = CliRunner().invoke(download_models)

        # Check that the command executed successfully
        assert result.exit_code == 0

        # Verify that download_models was called with the correct parameters
        mock_download.assert_called_once_with(
            progress=True,
            output_dir=DOCLING_CACHE_DIR,
        )


@pytest.mark.skip("Integration test")
def test_download_models_integration():
    """Integration test for the download-models command.

    This test actually downloads the models (to a temporary directory)
    and verifies that files were created.
    """

    # Create a temporary directory for the test
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Mock the DOCLING_CACHE_DIR to use our temporary directory
        with patch("pgai.vectorizer.parsing.DOCLING_CACHE_DIR", temp_path):
            from pgai.cli import download_models

            # Ensure directory is empty
            if temp_path.exists():
                for item in temp_path.iterdir():
                    if item.is_file():
                        item.unlink()
                    elif item.is_dir():
                        shutil.rmtree(item)

            # Run the CLI command
            result = CliRunner().invoke(download_models)

            # Check that the command executed successfully
            assert result.exit_code == 0

            # Verify that files were downloaded (directory should no longer be empty)
            assert any(
                temp_path.iterdir()
            ), "No files were downloaded to the models directory"



import logging
from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator
from typing import (
    Any,
    Literal,
    TextIO,
)

import psycopg
import yaml
from psycopg.rows import DictRow, dict_row
from psycopg.sql import SQL, Identifier
from pydantic import BaseModel, Field

from pgai.semantic_catalog.models import ObjectDescription

logger = logging.getLogger(__name__)

SCHEMA_VERSION = "1"


class Header(BaseModel):
    type: Literal["header"] = "header"
    schema_version: str = SCHEMA_VERSION

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class Column(BaseModel):
    name: str
    description: str


class Table(BaseModel):
    schema_name: str = Field(alias="schema")
    name: str
    type: Literal["table"] = "table"
    description: str
    columns: list[Column] = []

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True, by_alias=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class View(BaseModel):
    schema_name: str = Field(alias="schema")
    name: str
    type: Literal["view"] = "view"
    description: str
    columns: list[Column] = []

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True, by_alias=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class Function(BaseModel):
    schema_name: str = Field(alias="schema")
    name: str
    args: list[str] = []
    type: Literal["function"] = "function"
    description: str

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True, by_alias=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class Procedure(BaseModel):
    schema_name: str = Field(alias="schema")
    name: str
    args: list[str] = []
    type: Literal["procedure"] = "procedure"
    description: str

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True, by_alias=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class Aggregate(BaseModel):
    schema_name: str = Field(alias="schema")
    name: str
    args: list[str] = []
    type: Literal["aggregate"] = "aggregate"
    description: str

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True, by_alias=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class Fact(BaseModel):
    type: Literal["fact"] = "fact"
    id: int | None = None
    description: str

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


class SQLExample(BaseModel):
    type: Literal["sql_example"] = "sql_example"
    id: int | None = None
    sql: str
    description: str

    def to_yaml(self) -> str:
        return yaml.safe_dump(
            self.model_dump(exclude_none=True),
            explicit_start=True,
            sort_keys=False,
            explicit_end=True,
        )


Item = Table | View | Procedure | Function | Aggregate | SQLExample | Fact


def item_from_dict(d: dict[str, Any]) -> Item:
    """Convert a dictionary to the appropriate catalog item type.

    Takes a dictionary that represents a catalog item and converts it to the
    corresponding model class based on the 'type' field.

    Args:
        d: Dictionary containing item data with a 'type' field.

    Returns:
        An instance of the appropriate Item type.

    Raises:
        ValueError: If the 'type' field is not recognized.
    """
    mapping = {
        "table": Table,
        "view": View,
        "procedure": Procedure,
        "function": Function,
        "aggregate": Aggregate,
        "sql_example": SQLExample,
        "fact": Fact,
    }
    item_type = d.get("type")
    if item_type not in mapping:
        raise ValueError(f"unrecognized type: {item_type}")
    return mapping[item_type].model_validate(d)


def import_from_yaml(text: TextIO) -> Generator[Item, None, None]:
    """Import catalog items from a YAML file.

    Reads a YAML stream containing multiple documents. The first document should be a
    header with schema version information, followed by one or more catalog items.

    Args:
        text: A text IO stream containing YAML documents.

    Returns:
        A generator yielding catalog items (Tables, Views, Functions, etc.).

    Raises:
        RuntimeError: If the first document is not a header or has an invalid schema version.
    """
    for i, doc in enumerate(yaml.safe_load_all(text)):
        if not doc:
            continue
        if i == 0:
            schema_version = doc.get("schema_version")
            if schema_version is None:
                raise RuntimeError("the first document in yaml must be a header")
            if schema_version != SCHEMA_VERSION:
                raise RuntimeError(f"invalid schema_version: {schema_version}")
            continue
        yield item_from_dict(doc)


async def _look_up_class(
    target_con: psycopg.AsyncConnection, item: Table | View
) -> list[ObjectDescription]:
    """Look up database object IDs for a table or view and its columns.

    Queries the database to find the object IDs for a table or view and all its columns,
    then creates ObjectDescription objects for each.

    Args:
        target_con: Asynchronous database connection to the target database.
        item: A Table or View object to look up.

    Returns:
        A list of ObjectDescription objects for the table/view and its columns.

    Raises:
        RuntimeError: If the table or view is not found in the database.
    """
    descs: list[ObjectDescription] = []
    async with target_con.cursor(row_factory=dict_row) as cur:
        await cur.execute(
            """\
            select x.classid, x.objid
            from pg_get_object_address
            ( %(type)s
            , array[%(schema_name)s, %(name)s]::text[]
            , array[]::text[]
            ) x
        """,
            dict(
                type=item.type,
                schema_name=item.schema_name,
                name=item.name,
            ),
        )
        row = await cur.fetchone()
        if not row:
            raise RuntimeError(f"table/view not found: {item.schema_name}.{item.name}")
        classid = row["classid"]
        objid = row["objid"]
        descs.append(
            ObjectDescription(
                classid=classid,
                objid=objid,
                objsubid=0,
                objtype=item.type,
                objnames=[item.schema_name, item.name],
                objargs=[],
                description=item.description,
            )
        )
        await cur.execute(
            """\
            select c, x.objsubid
            from unnest(%(columns)s::text[]) c
            cross join lateral pg_get_object_address
            ( 'table column'
            , array[%(schema_name)s, %(name)s, c]::text[]
            , array[]::text[]
            ) x
        """,
            dict(
                type=item.type,
                schema_name=item.schema_name,
                name=item.name,
                columns=[c.name for c in item.columns],
            ),
        )
        name_to_id: dict[str, int] = {}
        for row in await cur.fetchall():
            name_to_id[row["c"]] = row["objsubid"]
        for column in item.columns:
            descs.append(
                ObjectDescription(
                    classid=classid,
                    objid=objid,
                    objsubid=name_to_id[column.name],
                    objtype=f"{item.type} column",
                    objnames=[item.schema_name, item.name, column.name],
                    objargs=[],
                    description=column.description,
                )
            )
    return descs


async def _look_up_proc(
    target_con: psycopg.AsyncConnection, item: Procedure | Function | Aggregate
) -> ObjectDescription:
    """Look up database object IDs for a procedure, function, or aggregate.

    Queries the database to find the object ID for a procedure, function, or aggregate,
    then creates an ObjectDescription object for it.

    Args:
        target_con: Asynchronous database connection to the target database.
        item: A Procedure, Function, or Aggregate object to look up.

    Returns:
        An ObjectDescription object for the procedure, function, or aggregate.

    Raises:
        RuntimeError: If the procedure, function, or aggregate is not found in the database.
    """
    async with target_con.cursor(row_factory=dict_row) as cur:
        await cur.execute(
            """\
            select x.classid, x.objid
            from pg_get_object_address
            ( %(type)s
            , array[%(schema_name)s, %(name)s]::text[]
            , %(args)s::text[]
            ) x
        """,
            dict(
                type=item.type,
                schema_name=item.schema_name,
                name=item.name,
                args=item.args,
            ),
        )
        row = await cur.fetchone()
        if not row:
            raise RuntimeError(
                f"procedure/function not found: {item.schema_name}.{item.name}"
            )
        return ObjectDescription(
            classid=row["classid"],
            objid=row["objid"],
            objsubid=0,
            objtype=item.type,
            objnames=[item.schema_name, item.name],
            objargs=item.args,
            description=item.description,
        )


async def _save_objs(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, objs: list[ObjectDescription]
) -> None:
    """Save object descriptions to the semantic catalog.

    Stores the provided object descriptions in the semantic catalog database.

    Args:
        catalog_con: Asynchronous database connection to the catalog database.
        catalog_id: ID of the semantic catalog to save to.
        objs: List of ObjectDescription objects to save.
    """
    async with catalog_con.cursor() as cur:
        for obj in objs:
            await cur.execute(
                """\
                select ai.sc_set_obj_desc
                ( %(classid)s
                , %(objid)s
                , %(objsubid)s
                , %(objtype)s
                , %(objnames)s
                , %(objargs)s
                , %(description)s
                , c.catalog_name
                )
                from ai.semantic_catalog c
                where c.id = %(catalog_id)s
            """,
                dict(
                    catalog_id=catalog_id,
                    classid=obj.classid,
                    objid=obj.objid,
                    objsubid=obj.objsubid,
                    objtype=obj.objtype,
                    objnames=obj.objnames,
                    objargs=obj.objargs,
                    description=obj.description,
                ),
            )


async def _save_sql_example(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, sql_example: SQLExample
) -> None:
    """Save a SQL example to the semantic catalog.

    Stores the provided SQL example in the semantic catalog database.

    Args:
        catalog_con: Asynchronous database connection to the catalog database.
        catalog_id: ID of the semantic catalog to save to.
        sql_example: SQLExample object to save.
    """
    async with catalog_con.cursor() as cur:
        if sql_example.id:
            await cur.execute(
                """\
                select ai.sc_update_sql_desc
                ( %(id)s
                , %(sql)s
                , %(description)s
                , c.catalog_name
                )
                from ai.semantic_catalog c
                where c.id = %(catalog_id)s
            """,
                dict(
                    catalog_id=catalog_id,
                    id=sql_example.id,
                    sql=sql_example.sql,
                    description=sql_example.description,
                ),
            )
        else:
            await cur.execute(
                """\
                select ai.sc_add_sql_desc
                ( %(sql)s
                , %(description)s
                , c.catalog_name
                )
                from ai.semantic_catalog c
                where c.id = %(catalog_id)s
            """,
                dict(
                    catalog_id=catalog_id,
                    sql=sql_example.sql,
                    description=sql_example.description,
                ),
            )


async def _save_fact(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, fact: Fact
) -> None:
    """Save a fact to the semantic catalog.

    Stores the provided fact in the semantic catalog database.

    Args:
        catalog_con: Asynchronous database connection to the catalog database.
        catalog_id: ID of the semantic catalog to save to.
        fact: Fact object to save.
    """
    async with catalog_con.cursor() as cur:
        if fact.id:
            await cur.execute(
                """\
                select ai.sc_update_fact
                ( %(id)s
                , %(description)s
                , c.catalog_name
                )
                from ai.semantic_catalog c
                where c.id = %(catalog_id)s
            """,
                dict(catalog_id=catalog_id, id=fact.id, description=fact.description),
            )
        else:
            await cur.execute(
                """\
                select ai.sc_add_fact
                ( %(description)s
                , c.catalog_name
                )
                from ai.semantic_catalog c
                where c.id = %(catalog_id)s
            """,
                dict(catalog_id=catalog_id, description=fact.description),
            )


async def save_to_catalog(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    items: Iterator[Item],
) -> None:
    """Save catalog items to the semantic catalog database.

    Takes a collection of catalog items and saves them to the semantic catalog database.
    Resolves object IDs from the target database and creates appropriate descriptions.

    Args:
        catalog_con: Asynchronous database connection to the catalog database.
        target_con: Asynchronous database connection to the target database (where objects are defined).
        catalog_id: ID of the semantic catalog to save to.
        items: Iterator of catalog items to save.
    """
    for item in items:
        match item.type:
            case "table" | "view":
                await _save_objs(
                    catalog_con, catalog_id, await _look_up_class(target_con, item)
                )
            case "function" | "procedure" | "aggregate":
                await _save_objs(
                    catalog_con, catalog_id, [await _look_up_proc(target_con, item)]
                )
            case "sql_example":
                await _save_sql_example(catalog_con, catalog_id, item)
            case "fact":
                await _save_fact(catalog_con, catalog_id, item)


async def _load_tables_views(
    cur: psycopg.AsyncCursor[DictRow], catalog_id: int
) -> AsyncGenerator[Item, None]:
    """Load table and view descriptions from the semantic catalog.

    Queries the semantic catalog to retrieve table and view descriptions,
    including column descriptions.

    Args:
        cur: Asynchronous database cursor.
        catalog_id: ID of the semantic catalog to load from.

    Returns:
        An async generator yielding Table and View objects.
    """
    await cur.execute(
        SQL("""\
            select
              x.objnames[1] as schema
            , x.objnames[2] as name
            , x.objtype as type
            , x.description
            , (
                select jsonb_agg
                (
                    jsonb_build_object
                    ( 'name', x1.objnames[3]
                    , 'description', x1.description
                    )
                    order by x1.objsubid
                )
                from ai.{} x1
                where x.classid = x1.classid
                and x.objid = x1.objid
                and x1.objsubid != 0
              ) as columns
            from ai.{} x
            where x.objtype in ('table', 'view')
        """).format(
            Identifier(f"semantic_catalog_obj_{catalog_id}"),
            Identifier(f"semantic_catalog_obj_{catalog_id}"),
        )
    )
    for row in await cur.fetchall():
        yield item_from_dict(row)


async def _load_functions_procedures_aggregates(
    cur: psycopg.AsyncCursor[DictRow], catalog_id: int
) -> AsyncGenerator[Item, None]:
    """Load function, procedure, and aggregate descriptions from the semantic catalog.

    Queries the semantic catalog to retrieve function, procedure, and aggregate descriptions.

    Args:
        cur: Asynchronous database cursor.
        catalog_id: ID of the semantic catalog to load from.

    Returns:
        An async generator yielding Function, Procedure, and Aggregate objects.
    """
    await cur.execute(
        SQL("""\
            select
              x.objnames[1] as schema
            , x.objnames[2] as name
            , x.objargs as args
            , x.objtype as type
            , x.description
            from ai.{} x
            where x.objtype in ('function', 'procedure', 'aggregate')
        """).format(Identifier(f"semantic_catalog_obj_{catalog_id}"))
    )
    for row in await cur.fetchall():
        yield item_from_dict(row)


async def _load_sql_examples(
    cur: psycopg.AsyncCursor[DictRow], catalog_id: int
) -> AsyncGenerator[Item, None]:
    """Load SQL examples from the semantic catalog.

    Queries the semantic catalog to retrieve SQL examples.

    Args:
        cur: Asynchronous database cursor.
        catalog_id: ID of the semantic catalog to load from.

    Returns:
        An async generator yielding SQLExample objects.
    """
    await cur.execute(
        SQL("""\
            select
              x.id
            , x.sql
            , x.description
            from ai.{} x
        """).format(Identifier(f"semantic_catalog_sql_{catalog_id}"))
    )
    for row in await cur.fetchall():
        row["type"] = "sql_example"
        yield item_from_dict(row)


async def _load_facts(
    cur: psycopg.AsyncCursor[DictRow], catalog_id: int
) -> AsyncGenerator[Item, None]:
    """Load facts from the semantic catalog.

    Queries the semantic catalog to retrieve facts.

    Args:
        cur: Asynchronous database cursor.
        catalog_id: ID of the semantic catalog to load from.

    Returns:
        An async generator yielding Fact objects.
    """
    await cur.execute(
        SQL("""\
            select
              x.id
            , x.description
            from ai.{} x
        """).format(Identifier(f"semantic_catalog_fact_{catalog_id}"))
    )
    for row in await cur.fetchall():
        row["type"] = "fact"
        yield item_from_dict(row)


async def load_from_catalog(
    con: psycopg.AsyncConnection, catalog_id: int
) -> AsyncGenerator[Item, None]:
    """Load all items from the semantic catalog.

    Retrieves all catalog items (tables, views, functions, procedures, aggregates,
    SQL examples, and facts) from the specified semantic catalog.

    Args:
        con: Asynchronous database connection to the catalog database.
        catalog_id: ID of the semantic catalog to load from.

    Returns:
        An async generator yielding all catalog items.
    """
    async with con.cursor(row_factory=dict_row) as cur:
        async for item in _load_tables_views(cur, catalog_id):
            yield item
        async for item in _load_functions_procedures_aggregates(cur, catalog_id):
            yield item
        async for item in _load_sql_examples(cur, catalog_id):
            yield item
        async for item in _load_facts(cur, catalog_id):
            yield item


def export_to_yaml(text: TextIO, items: Iterator[Item]) -> None:
    """Export catalog items to a YAML file.

    Writes catalog items to a YAML file with multiple documents. The first document
    is a header with schema version information, followed by one document per item.

    Args:
        text: A text IO stream to write the YAML documents to.
        items: Iterator of catalog items to export.
    """
    text.write(Header().to_yaml())
    for item in items:
        text.write(item.to_yaml())
        text.flush()


async def async_export_to_yaml(text: TextIO, items: AsyncIterator[Item]) -> None:
    """Export catalog items to a YAML file asynchronously.

    Asynchronously writes catalog items to a YAML file with multiple documents. The first document
    is a header with schema version information, followed by one document per item.

    Args:
        text: A text IO stream to write the YAML documents to.
        items: Async iterator of catalog items to export.
    """
    text.write(Header().to_yaml())
    async for item in items:
        text.write(item.to_yaml())
        text.flush()



import os
from datetime import datetime
from typing import Literal

import dotenv
import psycopg
from fastapi import FastAPI, Header, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field

dotenv.load_dotenv()

app = FastAPI()

DB_URL = "postgres://postgres@127.0.0.1:5432/test"


class ChunkingCharacterTextSplitter(BaseModel):
    implementation: Literal["character_text_splitter"]
    config_type: Literal["chunking"]
    chunk_size: int
    chunk_overlap: int
    separator: str | None
    is_separator_regex: bool | None


class ChunkingRecursiveCharacterTextSplitter(BaseModel):
    implementation: Literal["recursive_character_text_splitter"]
    config_type: Literal["chunking"]
    chunk_size: int
    chunk_overlap: int
    separators: list[str] | None
    is_separator_regex: bool | None


class ChunkingNone(BaseModel):
    implementation: Literal["none"]
    config_type: Literal["chunking"]


class EmbeddingOpenAI(BaseModel):
    implementation: Literal["openai"]
    config_type: Literal["embedding"]
    model: str
    dimensions: int
    user: str | None = None
    api_key_name: str | None = "OPENAI_API_KEY"


class FormattingPythonTemplate(BaseModel):
    implementation: Literal["python_template"]
    config_type: Literal["formatting"]
    template: str


class IndexingNone(BaseModel):
    implementation: Literal["none"]
    config_type: Literal["indexing"]


class IndexingDiskANN(BaseModel):
    implementation: Literal["diskann"]
    config_type: Literal["indexing"]
    min_rows: int
    storage_layout: Literal["memory_optimized"] | Literal["plain"] | None = None
    num_neighbors: int | None = None
    search_list_size: int | None = None
    max_alpha: float | None = None
    num_dimensions: int | None = None
    num_bits_per_dimension: int | None = None


class IndexingHNSW(BaseModel):
    implementation: Literal["hnsw"]
    config_type: Literal["indexing"]
    min_rows: int
    opclass: (
        Literal["vector_ip_ops"]
        | Literal["vector_cosine_ops"]
        | Literal["vector_l1_ops"]
        | None
    ) = None
    m: int | None = None
    ef_construction: int | None = None


class SchedulingNone(BaseModel):
    implementation: Literal["none"]
    config_type: Literal["scheduling"]


class SchedulingPgCron(BaseModel):
    implementation: Literal["pg_cron"]
    config_type: Literal["scheduling"]
    schedule: str


class SchedulingTimescaledb(BaseModel):
    implementation: Literal["timescaledb"]
    config_type: Literal["scheduling"]
    schedule_interval: str | None = None
    initial_start: datetime | None = None
    fixed_schedule: bool | None = None
    timezone: str | None = None


class ProcessingDefault(BaseModel):
    implementation: Literal["default"]
    config_type: Literal["processing"]
    batch_size: int | None = None
    concurrency: int | None = None


class LoadingColumn(BaseModel):
    implementation: Literal["column"]
    config_type: Literal["loading"]
    column_name: str


class LoadingUri(BaseModel):
    implementation: Literal["uri"]
    config_type: Literal["loading"]
    column_name: str


class ParsingAuto(BaseModel):
    implementation: Literal["auto"]
    config_type: Literal["parsing"]


class ParsingNone(BaseModel):
    implementation: Literal["none"]
    config_type: Literal["parsing"]


class ParsingPyMuPDF(BaseModel):
    implementation: Literal["pymupdf"]
    config_type: Literal["parsing"]


class ParsingDocling(BaseModel):
    implementation: Literal["docling"]
    config_type: Literal["parsing"]


class Config(BaseModel):
    version: str
    indexing: IndexingNone | IndexingDiskANN | IndexingHNSW = Field(
        ..., discriminator="implementation"
    )
    formatting: FormattingPythonTemplate
    embedding: EmbeddingOpenAI
    scheduling: SchedulingNone | SchedulingPgCron | SchedulingTimescaledb = Field(
        ..., discriminator="implementation"
    )
    chunking: (
        ChunkingCharacterTextSplitter
        | ChunkingRecursiveCharacterTextSplitter
        | ChunkingNone
    ) = Field(..., discriminator="implementation")
    processing: ProcessingDefault = Field(..., discriminator="implementation")
    loading: LoadingColumn | LoadingUri = Field(..., discriminator="implementation")
    parsing: ParsingAuto | ParsingNone | ParsingPyMuPDF | ParsingDocling = Field(
        ..., discriminator="implementation"
    )


class PrimaryKeyColumn(BaseModel):
    attnum: int
    pknum: int
    attname: str
    typname: str


class Vectorizer(BaseModel):
    id: int
    source_schema: str
    source_table: str
    source_pk: list[PrimaryKeyColumn]
    trigger_name: str
    queue_schema: str | None = None
    queue_table: str | None = None
    config: Config


def vectorize(v: Vectorizer) -> int:
    # pretend to work the queue
    with psycopg.connect(DB_URL) as con:
        while True:
            with con.cursor() as cur:
                cur.execute(f"""
                    delete from {v.queue_schema}.{v.queue_table}
                """)
                return cur.rowcount


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    return JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        content=jsonable_encoder({"detail": exc.errors(), "body": exc.body}),
    )


@app.post("/api/v1/events")
async def execute_vectorizer(vectorizer: Vectorizer):
    print(f"execute vectorizer: {vectorizer.id}")
    # do this in a blocking manner to make the tests easier
    # we KNOW that when the HTTP request has returned that the work has been done
    deleted = vectorize(vectorizer)
    print(f"queue emptied: {deleted} rows deleted")
    return {"id": vectorizer.id}


@app.get("/api/v1/projects/secrets")
async def get_secrets(secret_name: str = Header(None, alias="Secret-Name")):
    if not secret_name:
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={"error": "Secret-Name header is required"},
        )

    # For now, we'll just return the test key if the secret_name matches
    if secret_name == "OPENAI_API_KEY" or secret_name == "OPENAI_API_KEY_2":
        return {secret_name: "test"}
    elif secret_name == "OPENAI_API_KEY_REAL":
        return {secret_name: os.environ["OPENAI_API_KEY"]}
    elif secret_name == "COHERE_API_KEY_REAL":
        return {secret_name: os.environ["COHERE_API_KEY"]}
    elif secret_name == "ANTHROPIC_API_KEY_REAL":
        return {secret_name: os.environ["ANTHROPIC_API_KEY"]}
    elif secret_name == "ERROR_SECRET":
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={"error": "error secret"},
        )
    else:
        return JSONResponse(
            status_code=status.HTTP_404_NOT_FOUND,
            content={"error": f"Secret '{secret_name}' not found"},
        )



import os
from pathlib import Path

import pytest
from psycopg import Connection
from psycopg.rows import dict_row
from testcontainers.ollama import OllamaContainer  # type: ignore
from testcontainers.postgres import PostgresContainer  # type: ignore

from tests.vectorizer.cli.conftest import (
    TestDatabase,
    configure_vectorizer,
    run_vectorizer_worker,
    setup_source_table,
)


@pytest.fixture(scope="session")
def ollama_url():
    # If the OLLAMA_HOST environment variable is set, we assume that the user
    # has an Ollama container running and we don't need to start a new one.
    if "OLLAMA_HOST" in os.environ:
        yield os.environ["OLLAMA_HOST"]
    else:
        with OllamaContainer(
            image="ollama/ollama:latest",
            # Passing the ollama_home lets us reuse models that have already
            # been pulled to the `~/.ollama` path on the host machine.
            ollama_home=Path.home() / ".ollama",
        ) as ollama:
            yield ollama.get_endpoint()


def configure_ollama_vectorizer(
    connection: Connection,
    ollama_url: str,
    number_of_rows: int = 1,
    concurrency: int = 1,
    batch_size: int = 1,
    chunking: str = "chunking_character_text_splitter()",
    formatting: str = "formatting_python_template('$chunk')",
) -> int:
    """Creates and configures an ollama vectorizer for testing"""

    table_name = setup_source_table(connection, number_of_rows)
    return configure_vectorizer(
        table_name,
        connection,
        concurrency=concurrency,
        batch_size=batch_size,
        chunking=chunking,
        formatting=formatting,
        embedding=f"embedding_ollama('nomic-embed-text',"
        f"768, base_url => '{ollama_url}')",
    )


@pytest.mark.parametrize(
    "num_items,concurrency,batch_size",
    [
        (1, 1, 1),
        (4, 2, 2),
    ],
)
def test_ollama_vectorizer(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    ollama_url: str,
    num_items: int,
    concurrency: int,
    batch_size: int,
):
    """Test successful processing of vectorizer tasks"""
    _, conn = cli_db

    vectorizer_id = configure_ollama_vectorizer(
        cli_db[1],
        ollama_url,
        number_of_rows=num_items,
        concurrency=concurrency,
        batch_size=batch_size,
    )
    # Insert pre-existing embedding for first item
    with conn.cursor() as cur:
        cur.execute("""
           INSERT INTO
           blog_embedding_store(embedding_uuid, id, chunk_seq, chunk, embedding)
           VALUES (gen_random_uuid(), 1, 1, 'post_1',
            array_fill(0, ARRAY[768])::vector)
        """)
    result = run_vectorizer_worker(cli_db_url, vectorizer_id, concurrency)

    assert not result.exception
    assert result.exit_code == 0

    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) as count FROM blog_embedding_store;")
        assert cur.fetchone()["count"] == num_items  # type: ignore


@pytest.mark.parametrize(
    "chunking",
    [
        "chunking_character_text_splitter()",
        "chunking_recursive_character_text_splitter()",
    ],
)
def test_vectorization_successful_with_null_contents(
    cli_db: tuple[PostgresContainer, Connection],
    cli_db_url: str,
    ollama_url: str,
    chunking: str,
):
    _, conn = cli_db
    vectorizer_id = configure_ollama_vectorizer(conn, ollama_url, chunking=chunking)
    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("ALTER TABLE blog ALTER COLUMN content DROP NOT NULL;")
        cur.execute("UPDATE blog SET content = null;")

    result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    assert not result.exception
    assert result.exit_code == 0

    _, conn = cli_db

    with conn.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) as count FROM blog_embedding_store;")
        assert cur.fetchone()["count"] == 0  # type: ignore



import io
import logging

import psycopg
from psycopg.sql import SQL, Composed, Identifier, Literal

logger = logging.getLogger(__name__)


async def _sample_as_inserts(
    con: psycopg.AsyncConnection, schema_name: str, object_name: str, limit: int = 3
) -> str:
    """Generate INSERT statements for sample data from a database object.

    Queries the specified database object and generates INSERT statements
    for a limited number of rows.

    Args:
        con: Asynchronous database connection object.
        schema_name: Name of the schema containing the object.
        object_name: Name of the table or view to sample.
        limit: Maximum number of rows to sample (default: 3).

    Returns:
        A string containing SQL INSERT statements for the sampled data.
    """
    async with con.cursor() as cur:
        await cur.execute(
            SQL("select * from {}.{} limit %s").format(
                Identifier(schema_name), Identifier(object_name)
            ),
            (limit,),
        )
        assert cur.description is not None, "cursor description not set"
        rows: list[Composed] = []
        for row in await cur.fetchall():
            rows.append(SQL("({})").format(SQL(", ").join(Literal(val) for val in row)))
        logger.debug(f"sampled {len(rows)} rows from {schema_name}.{object_name}")
        sql = SQL("INSERT INTO {}.{} ({}) VALUES\n  {};").format(
            Identifier(schema_name),
            Identifier(object_name),
            SQL(", ").join([Identifier(col.name) for col in cur.description]),
            SQL("\n, ").join(rows),
        )
        return sql.as_string(con)


async def _sample_as_copy_text(
    con: psycopg.AsyncConnection, schema_name: str, object_name: str, limit: int = 3
) -> str:
    """Generate COPY command text for sample data from a database object.

    Queries the specified database object and generates a COPY command with
    sample data in text format, including headers.

    Args:
        con: Asynchronous database connection object.
        schema_name: Name of the schema containing the object.
        object_name: Name of the table or view to sample.
        limit: Maximum number of rows to sample (default: 3).

    Returns:
        A string containing the COPY command followed by the sampled data in text format.
    """
    query = SQL(
        "COPY (SELECT * FROM {}.{} LIMIT {}) TO STDOUT WITH (FORMAT TEXT, HEADER true)"
    ).format(
        Identifier(schema_name),
        Identifier(object_name),
        Literal(limit),
    )
    buf = io.StringIO()
    buf.write(query.as_string(con))
    buf.write(";\n/*\n")

    async with con.cursor() as cur:
        copy_writer = io.StringIO()
        async with cur.copy(query) as copy_:
            while data := await copy_.read():
                copy_writer.write(bytes(data).decode("utf-8"))
        sample_data = copy_writer.getvalue()
        buf.write(sample_data)

    buf.write("*/")
    return buf.getvalue()


async def sample_table(
    con: psycopg.AsyncConnection,
    schema_name: str,
    table_name: str,
    limit: int = 3,
    format: str = "copy_text",
) -> str:
    """Sample data from a database table.

    Retrieves a limited number of rows from the specified table and formats the data
    according to the specified format.

    Args:
        con: Asynchronous database connection object.
        schema_name: Name of the schema containing the table.
        table_name: Name of the table to sample.
        limit: Maximum number of rows to sample (default: 3).
        format: Output format, either "inserts" or "copy_text" (default: "copy_text").

    Returns:
        A string containing the sampled data in the requested format.

    Raises:
        RuntimeError: If an unsupported format is specified.
    """
    assert format in {"inserts", "copy_text"}
    match format:
        case "inserts":
            return await _sample_as_inserts(con, schema_name, table_name, limit)
        case "copy_text":
            return await _sample_as_copy_text(con, schema_name, table_name, limit)
        case _:
            raise RuntimeError(f"unsupported format: {format}")


async def sample_view(
    con: psycopg.AsyncConnection,
    schema_name: str,
    view_name: str,
    limit: int = 3,
    format: str = "copy_text",
) -> str:
    """Sample data from a database view.

    Retrieves a limited number of rows from the specified view and formats the data
    according to the specified format.

    Args:
        con: Asynchronous database connection object.
        schema_name: Name of the schema containing the view.
        view_name: Name of the view to sample.
        limit: Maximum number of rows to sample (default: 3).
        format: Output format, either "inserts" or "copy_text" (default: "copy_text").

    Returns:
        A string containing the sampled data in the requested format.

    Raises:
        RuntimeError: If an unsupported format is specified.
    """
    assert format in {"inserts", "copy_text"}
    match format:
        case "inserts":
            return await _sample_as_inserts(con, schema_name, view_name, limit)
        case "copy_text":
            return await _sample_as_copy_text(con, schema_name, view_name, limit)
        case _:
            raise RuntimeError(f"unsupported format: {format}")



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel
from typing_extensions import override

from pgai.vectorizer.loading import LoadedDocument


class ParsingNone(BaseModel):
    implementation: Literal["none"]

    def parse(self, _1: dict[str, Any], payload: str | LoadedDocument) -> str:  # noqa: ARG002
        if isinstance(payload, LoadedDocument):
            raise ValueError(
                "Cannot chunk Document with parsing_none, "
                "use parsing_auto or parsing_pymupdf"
            )
        return payload


class ParsingAuto(BaseModel):
    implementation: Literal["auto"]

    def parse(self, row: dict[str, Any], payload: str | LoadedDocument) -> str:
        if isinstance(payload, LoadedDocument):
            if payload.file_type == "epub":
                # epub is not supported by docling, but by pymupdf
                return ParsingPyMuPDF(implementation="pymupdf").parse(row, payload)

            return ParsingDocling(implementation="docling").parse(row, payload)
        else:
            return payload


class BaseDocumentParsing(BaseModel, ABC):
    """Base class for document parsing implementations."""

    implementation: str

    def parse(self, row: dict[str, Any], payload: LoadedDocument | str) -> str:
        """
        Parse a document payload into a string representation.

        Args:
            row: Metadata about the document. The whole actual db row
            payload: Either a LoadedDocument or raw string content

        Returns:
            Parsed string content. Markdown preferable.

        Raises:
            ValueError: If payload type is invalid or file type cannot be determined
        """
        if isinstance(payload, str):
            raise ValueError(
                f"Column content must be a document to be parsed by "
                f"{self.implementation}, use parsing_auto or parsing_none instead"
            )

        if payload.file_type is None:
            raise ValueError("No file extension could be determined")

        if payload.file_type in ["txt", "md"]:
            return payload.content.getvalue().decode("utf-8")

        return self.parse_doc(row, payload)

    @abstractmethod
    def parse_doc(self, row: dict[str, Any], payload: LoadedDocument) -> str:
        """
        Parse a binary document into a string representation, Markdown preferable.
        Must be implemented by subclasses.
        """


class ParsingPyMuPDF(BaseDocumentParsing):
    """Document parsing implementation using PyMuPDF."""

    implementation: Literal["pymupdf"]  # type: ignore[reportIncompatibleVariableOverride]

    @override
    def parse_doc(self, row: dict[str, Any], payload: LoadedDocument) -> str:  # noqa: ARG002
        # Note: deferred import to avoid import overhead
        import pymupdf  # type: ignore
        import pymupdf4llm  # type: ignore

        with pymupdf.open(
            stream=payload.content, filetype=payload.file_type
        ) as pdf_document:  # type: ignore
            return pymupdf4llm.to_markdown(pdf_document)  # type: ignore


DEFAULT_CACHE_DIR = Path.home().joinpath(".cache/docling/models")
cache_dir = os.getenv("VECTORIZER_DOCLING_CACHE_DIR")
DOCLING_CACHE_DIR = DEFAULT_CACHE_DIR if cache_dir is None else Path(cache_dir)


class ParsingDocling(BaseDocumentParsing):
    """Document parsing implementation using Docling."""

    implementation: Literal["docling"]  # type: ignore[reportIncompatibleVariableOverride]
    cache_dir: Path | str = DOCLING_CACHE_DIR

    @override
    def parse_doc(self, row: dict[str, Any], payload: LoadedDocument) -> str:  # noqa: ARG002
        # Note: deferred import to avoid import overhead
        from docling.datamodel.base_models import (
            DocumentStream,  # type: ignore
            InputFormat,
        )
        from docling.datamodel.pipeline_options import PdfPipelineOptions
        from docling.document_converter import (
            DocumentConverter,
            ImageFormatOption,
            PdfFormatOption,
        )

        basic_pipeline_options = PdfPipelineOptions(
            do_ocr=False,  # we do not want to do OCR in PDF (yet)
            artifacts_path=self.cache_dir if os.path.isdir(self.cache_dir) else None,
        )  # pyright: ignore[reportCallIssue]

        with_ocr_pipeline_options = basic_pipeline_options
        with_ocr_pipeline_options.do_ocr = True

        converter = DocumentConverter(
            format_options={
                InputFormat.PDF: PdfFormatOption(
                    pipeline_options=basic_pipeline_options
                ),
                InputFormat.IMAGE: ImageFormatOption(
                    pipeline_options=with_ocr_pipeline_options
                ),
                # we don't need to configure the rest of the formats as they follow the
                # simple pipeline without external models, OCR, etc.
            }
        )

        source = DocumentStream(name=payload.file_path or "", stream=payload.content)
        result = converter.convert(source)
        return result.document.export_to_markdown()



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_processing_default():
    tests = [
        (
            "select ai.processing_default()",
            {
                "implementation": "default",
                "config_type": "processing",
            },
        ),
        (
            "select ai.processing_default(batch_size=>500)",
            {
                "implementation": "default",
                "config_type": "processing",
                "batch_size": 500,
            },
        ),
        (
            "select ai.processing_default(concurrency=>3)",
            {
                "implementation": "default",
                "config_type": "processing",
                "concurrency": 3,
            },
        ),
        (
            "select ai.processing_default(batch_size=>500, concurrency=>3)",
            {
                "implementation": "default",
                "config_type": "processing",
                "batch_size": 500,
                "concurrency": 3,
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_validate_processing():
    ok = [
        "select ai._validate_processing(ai.processing_default())",
        "select ai._validate_processing(ai.processing_default(batch_size=>500))",
        "select ai._validate_processing(ai.processing_default(batch_size=>2048))",
        "select ai._validate_processing(ai.processing_default(batch_size=>2048, concurrency=>1))",
        "select ai._validate_processing(ai.processing_default(concurrency=>10))",
    ]
    bad = [
        (
            "select ai._validate_processing(ai.indexing_hnsw())",
            "invalid config_type for processing config",
        ),
        (
            """select ai._validate_processing('{"config_type": "processing"}'::jsonb)""",
            "processing implementation not specified",
        ),
        (
            """
            select ai._validate_processing
            ( '{"config_type": "processing", "implementation": "grandfather clock"}'::jsonb
            )
            """,
            'unrecognized processing implementation: "grandfather clock"',
        ),
        (
            """
            select ai._validate_processing
            ( ai.processing_default(batch_size=>2049)
            )
            """,
            "batch_size must be less than or equal to 2048",
        ),
        (
            """
            select ai._validate_processing
            ( ai.processing_default(batch_size=>0)
            )
            """,
            "batch_size must be greater than 0",
        ),
        (
            """
            select ai._validate_processing
            ( ai.processing_default(concurrency=>0)
            )
            """,
            "concurrency must be greater than 0",
        ),
        (
            """
            select ai._validate_processing
            ( ai.processing_default(concurrency=>51)
            )
            """,
            "concurrency must be less than or equal to 50",
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            for query in ok:
                cur.execute(query)
                assert True
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")



from typing import Any

import pytest
from pydantic import BaseModel

from pgai.vectorizer.migrations import migrations as global_migrations
from pgai.vectorizer.migrations import register_migration
from pgai.vectorizer.vectorizer import Vectorizer

config_0_9_0 = {
    "version": "0.9.0",
    "chunking": {
        "separator": "\n\n",
        "chunk_size": 128,
        "config_type": "chunking",
        "chunk_column": "body",
        "chunk_overlap": 10,
        "implementation": "character_text_splitter",
        "is_separator_regex": False,
    },
    "indexing": {"config_type": "indexing", "implementation": "none"},
    "embedding": {
        "model": "text-embedding-3-small",
        "dimensions": 768,
        "config_type": "embedding",
        "api_key_name": "OPENAI_API_KEY",
        "implementation": "openai",
    },
    "formatting": {
        "template": "title: $title published: $published $chunk",
        "config_type": "formatting",
        "implementation": "python_template",
    },
    "processing": {"config_type": "processing", "implementation": "default"},
    "scheduling": {
        "job_id": 1000,
        "timezone": "America/Chicago",
        "config_type": "scheduling",
        "initial_start": "2050-01-06T00:00:00+00:00",
        "implementation": "timescaledb",
        "schedule_interval": "00:05:00",
    },
}


config_0_10_0 = {
    "version": "0.10.0",
    "loading": {
        "config_type": "loading",
        "implementation": "column",
        "retries": 6,
        "column_name": "body",
    },
    "parsing": {"config_type": "parsing", "implementation": "none"},
    "chunking": {
        "separator": "\n\n",
        "chunk_size": 128,
        "config_type": "chunking",
        "chunk_overlap": 10,
        "implementation": "character_text_splitter",
        "is_separator_regex": False,
    },
    "indexing": {"config_type": "indexing", "implementation": "none"},
    "embedding": {
        "model": "text-embedding-3-small",
        "dimensions": 768,
        "config_type": "embedding",
        "api_key_name": "OPENAI_API_KEY",
        "implementation": "openai",
    },
    "formatting": {
        "template": "title: $title published: $published $chunk",
        "config_type": "formatting",
        "implementation": "python_template",
    },
    "processing": {"config_type": "processing", "implementation": "default"},
    "scheduling": {
        "job_id": 1000,
        "timezone": "America/Chicago",
        "config_type": "scheduling",
        "initial_start": "2050-01-06T00:00:00+00:00",
        "implementation": "timescaledb",
        "schedule_interval": "00:05:00",
    },
    "destination": {
        "config_type": "destination",
        "implementation": "table",
        "target_schema": "public",
        "target_table": "target",
    },
}

vectorizer_fields = {
    "id": 1,
    "queue_schema": "public",
    "queue_table": "queue",
    "source_schema": "public",
    "source_table": "source",
    "target_schema": "public",
    "target_table": "target",
    "source_pk": [
        {
            "attname": "id",
            "pknum": 1,
            "attnum": 1,
        }
    ],
}


@pytest.fixture(autouse=True)
def clean_migrations():
    """
    Fixture to preserve and restore the global migrations list.
    This ensures that migrations registered in tests don't affect other tests.
    """
    # Save the original migrations list
    original_migrations = global_migrations.copy()

    # Yield control back to the test
    yield

    # Restore the original migrations list after the test
    global_migrations.clear()
    global_migrations.extend(original_migrations)


def test_multiple_migrations():
    class SimpleConfig(BaseModel):
        version: str

    class SimpleVectorizer(BaseModel):
        config: SimpleConfig

    # register some migrations
    def migration_func_1_2(old_vectorizer: SimpleVectorizer) -> dict[str, Any]:
        assert old_vectorizer.config.version == "0.0.1"
        old_vectorizer.config.version = "0.0.2"
        return old_vectorizer.model_dump()

    def migration_func_2_0_9(old_vectorizer: SimpleVectorizer) -> dict[str, Any]:
        assert old_vectorizer.config.version == "0.0.2"
        return {
            **vectorizer_fields,
            "config": config_0_9_0,
        }

    register_migration("0.0.2", SimpleVectorizer, "1 to 2")(migration_func_1_2)
    register_migration("0.9.0", SimpleVectorizer, "2 to 0.9.0")(migration_func_2_0_9)

    migreated_vectorizer = Vectorizer(
        **{  # type: ignore
            **vectorizer_fields,
            "config": config_0_10_0 | {"version": "0.0.1"},
        }
    )
    expected_vectorizer = Vectorizer(
        **{  # type: ignore
            **vectorizer_fields,
            "config": config_0_10_0 | {"original_version": "0.0.1"},
        }
    )

    assert migreated_vectorizer == expected_vectorizer


def test_migrate_config_from_ext_version_0_9_to_0_10():
    migrated_vectorizer = Vectorizer(
        **{  # type: ignore
            **vectorizer_fields,
            "config": config_0_9_0,
        }
    )  # pyright: ignore [reportArgumentType]
    expected_vectorizer = Vectorizer(
        **{  # type: ignore
            **vectorizer_fields,
            "config": config_0_10_0 | {"original_version": "0.9.0"},
        }
    )  # pyright: ignore [reportArgumentType]

    assert migrated_vectorizer == expected_vectorizer



import psycopg
import pytest
from psycopg.rows import dict_row

import pgai
import pgai.semantic_catalog as sc
from pgai.semantic_catalog.vectorizer import (
    OllamaConfig,
    OpenAIConfig,
    SentenceTransformersConfig,
)
from tests.semantic_catalog.utils import PostgresContainer

DATABASE = "sc_01"


@pytest.fixture(scope="module", autouse=True)
def database(container: PostgresContainer):
    container.drop_database(DATABASE, force=True)
    container.create_database(DATABASE)
    pgai.install(container.connection_string(database=DATABASE))


async def test_semantic_catalog(container: PostgresContainer):
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        # all defaults
        cat1 = await sc.create(con)
        assert cat1.name == "default"
        embs = await cat1.list_embeddings(con)
        assert len(embs) == 1
        emb_name, config = embs[0]
        assert emb_name == "emb1"
        assert isinstance(config, SentenceTransformersConfig)

        # add a second semantic catalog
        cat2 = await sc.create(con, catalog_name="cat2")
        assert cat2.name == "cat2"
        embs = await cat2.list_embeddings(con)
        assert len(embs) == 1
        emb_name, config = embs[0]
        assert emb_name == "emb1"
        assert isinstance(config, SentenceTransformersConfig)

        # add a third semantic catalog
        config = OllamaConfig.create(model="nomic-embed-text", dimensions=768)
        cat3 = await sc.create(
            con, catalog_name="cat3", embedding_name="vec1", embedding_config=config
        )
        assert cat3.name == "cat3"
        embs = await cat3.list_embeddings(con)
        assert len(embs) == 1
        emb_name, config = embs[0]
        assert emb_name == "vec1"
        assert isinstance(config, OllamaConfig)
        assert config.model == "nomic-embed-text"
        assert config.dimensions == 768

        # add a second embedding
        config = OpenAIConfig.create(model="text-embedding-ada-002", dimensions=1536)
        emb_name, config = await cat3.add_embedding(con, config)
        assert emb_name == "emb2"
        assert isinstance(config, OpenAIConfig)
        assert config.model == "text-embedding-ada-002"
        assert config.dimensions == 1536

        # add a third embedding
        config = SentenceTransformersConfig.create(
            model="nomic-ai/nomic-embed-text-v1.5", dimensions=768
        )
        emb_name, config = await cat3.add_embedding(con, config)
        assert emb_name == "emb3"
        assert isinstance(config, SentenceTransformersConfig)
        assert config.model == "nomic-ai/nomic-embed-text-v1.5"
        assert config.dimensions == 768

        # drop the second
        await cat3.drop_embedding(con, "emb2")
        embs = await cat3.list_embeddings(con)
        assert len(embs) == 2
        emb_name, config = embs[0]
        assert emb_name == "vec1"
        assert isinstance(config, OllamaConfig)
        assert config.model == "nomic-embed-text"
        assert config.dimensions == 768
        emb_name, config = embs[1]
        assert emb_name == "emb3"
        assert isinstance(config, SentenceTransformersConfig)
        assert config.model == "nomic-ai/nomic-embed-text-v1.5"
        assert config.dimensions == 768

        # add a fourth embedding
        config = SentenceTransformersConfig.create(
            model="nomic-ai/nomic-embed-text-v1.5", dimensions=768
        )
        emb_name, config = await cat3.add_embedding(con, config)
        assert emb_name == "emb4"
        assert isinstance(config, SentenceTransformersConfig)
        assert config.model == "nomic-ai/nomic-embed-text-v1.5"
        assert config.dimensions == 768

        async with con.cursor(row_factory=dict_row) as cur:
            # we should have 3 semantic catalogs
            await cur.execute("select count(*) as actual from ai.semantic_catalog")
            row = await cur.fetchone()
            assert row is not None
            actual: int = int(row["actual"])
            assert actual == 3
            # check how many embeddings each catalog has
            await cur.execute("""\
                select c.catalog_name, count(*) as num
                from ai.semantic_catalog_embedding e
                inner join ai.semantic_catalog c on (e.semantic_catalog_id = c.id)
                group by 1
            """)
            for row in await cur.fetchall():
                match row["catalog_name"]:
                    case "default" | "cat2":
                        assert row["num"] == 1
                    case "cat3":
                        assert row["num"] == 3
                    case _:
                        raise AssertionError("unexpected semantic catalog")
            # make sure the vector columns exist
            await cur.execute("""\
                select a.attnum, a.attname, format_type(a.atttypid, a.atttypmod) as type
                from pg_class k
                inner join pg_namespace n on (k.relnamespace = n.oid)
                inner join pg_attribute a on (k.oid = a.attrelid)
                where n.nspname = 'ai'
                and k.relname = 'semantic_catalog_obj_3'
                and a.attnum > 0
                and not a.attisdropped
                and format_type(a.atttypid, a.atttypmod) ~ '^vector\\([0-9]+\\)$'
                order by a.attnum
            """)
            for row in await cur.fetchall():
                match row["attnum"]:
                    case 10:
                        assert row["attname"] == "vec1"
                        assert row["type"] == "vector(768)"
                    case 12:
                        assert row["attname"] == "emb3"
                        assert row["type"] == "vector(768)"
                    case 13:
                        assert row["attname"] == "emb4"
                        assert row["type"] == "vector(768)"
                    case _:
                        raise AssertionError("unexpected vector column")

    # try to add a semantic catalog with a bad name
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        with pytest.raises(psycopg.errors.CheckViolation):
            await sc.create(con, catalog_name="MyCatalog")

    # add a semantic catalog with explicit embedding
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        config = OpenAIConfig.create(model="text-embedding-ada-002", dimensions=1536)
        cat1 = await sc.create(
            con,
            catalog_name="name_checks",
            embedding_name="vec1",
            embedding_config=config,
        )
        assert cat1.name == "name_checks"
        embs = await cat1.list_embeddings(con)
        assert len(embs) == 1
        emb_name, config = embs[0]
        assert emb_name == "vec1"
        assert isinstance(config, OpenAIConfig)

    # try to add a second embedding with a bad name
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        config = OpenAIConfig.create(model="text-embedding-ada-002", dimensions=1536)
        with pytest.raises(psycopg.errors.CheckViolation):
            await cat1.add_embedding(con, config, embedding_name="This is A bad name")

    # test dropping a semantic catalog
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        await cat1.drop(con)
        async with con.cursor(row_factory=dict_row) as cur:
            await cur.execute("select count(*) as actual from ai.semantic_catalog")
            row = await cur.fetchone()
            assert row is not None
            actual: int = int(row["actual"])
            assert actual == 3


async def test_semantic_catalog_getters(container: PostgresContainer):
    async with (
        await psycopg.AsyncConnection.connect(
            container.connection_string(database=DATABASE)
        ) as con,
        con.transaction(force_rollback=True) as _,
    ):
        cat1 = await sc.create(con, catalog_name="butch1")
        assert cat1.name == "butch1"
        cat2 = await sc.create(con, catalog_name="butch2")
        assert cat2.name == "butch2"
        cat3 = await sc.create(con, catalog_name="butch3")
        assert cat3.name == "butch3"
        actual = await sc.from_name(con, "butch2")
        assert actual.name == "butch2"
        assert actual.id == cat2.id
        actual = await sc.from_id(con, cat3.id)
        assert actual.name == "butch3"
        assert actual.id == cat3.id
        actuals = await sc.list_semantic_catalogs(con)
        assert len(actuals) >= 3
        assert "butch1" in {c.name for c in actuals}
        assert "butch2" in {c.name for c in actuals}
        assert "butch3" in {c.name for c in actuals}
